#!/usr/bin/env python

"""
########################################################################
# Author:	   Carlos A. Ruiz-Perez
# Email:       cruizperez3@gatech.edu
# Institution: Georgia Institute of Technology
# Version:	   0.9
# Date:		   14 February 2020

# Description: This script filters a fasta file based on a list of IDs
# to retain or exclude
########################################################################
"""

################################################################################
"""---1.0 Import Modules---"""
import argparse
from sys import argv

################################################################################
"""---2.0 Define Functions---"""

def fastA_filter_list(fasta_file, output_file, id_list, reverse=False):
    # import modules
    from Bio.SeqIO.FastaIO import SimpleFastaParser
    # Filter fasta file
    with open(fasta_file) as fasta_input, open(output_file, 'w') as output:
        if reverse == True:
            for title, seq in SimpleFastaParser(fasta_input):
                if title.split()[0] not in id_list:
                    output.write(">{}\n{}\n".format(title, seq))
        else:
            for title, seq in SimpleFastaParser(fasta_input):
                if len(id_list) < 1:
                    break
                elif title.split()[0] in id_list:
                    output.write(">{}\n{}\n".format(title, seq))
                    id_list.remove(title.split()[0])

def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
    description='''Filter a FastA file based ona provided list of IDs (file or input).\n'''
    '''It can exclude or retrieve the sequences using the --reverse flag\n'''
    '''Usage: ''' + argv[0] + ''' -f [FastA File] -o [output_file File] -l [File with IDs] OR -i [ID list]\n'''
    '''Global mandatory parameters: [FastA_File] [output_file_File] [ID List File]\n'''
    'Optional Database Parameters: See ' + argv[0] + ' -h')
    parser.add_argument('-f', '--fasta', dest='fasta_file', action='store', required=True, 
    help='FastA file to filter')
    parser.add_argument('-o', '--output_file', dest='output_file', action='store', required=True, 
    help='output_file FastA file with retrieved sequences')
    parser.add_argument('-l', '--list', dest='id_file', action='store', required=False, 
    help='File with list of IDs to filter.')
    parser.add_argument('-i', '--id', dest='id_list', action='store', required=False,
    help='Comma-separated IDs to filter: ID1,ID2,ID3')
    parser.add_argument('--reverse', action='store_true', 
    help='Exclude the sequences in the list file. By default False, i.e. retrieves those in the list')
    args = parser.parse_args()

    fasta_file = args.fasta_file
    output_file = args.output_file
    id_file = args.id_file
    id_list = args.id_list
    reverse = args.reverse

    if id_list != None:
        id_list = id_list.split(",")
        fastA_filter_list(fasta_file, output_file, id_list, reverse)
    elif id_file != None:
        id_list = []
        with open(id_file, 'r') as file:
            for line in file:
                id_list.append(line.strip())
        fastA_filter_list(fasta_file, output_file, id_list, reverse)
    else:
        raise ValueError("Did you provide the IDs to filter?")

if __name__ == "__main__":
    main()


#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz-Perez
# Email:        cruizperez3@gatech.edu
# Institution:  Georgia Institute of Technology
# Date:         11 April 2021

# Description: This script downloads the profiles and scores for all hmm models
# used in KOfamscan
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from collections import namedtuple
from collections import Counter
from datetime import date
from typing import Tuple
from typing import Dict
from pathlib import Path
from sys import argv

import subprocess
import argparse
import shutil
import wget
# ==============================================================================


# ==============================================================================
# TODO: Upload data to our server for backup
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Download latest KOfamscan profiles and scores
def kofam_downloader(outdir: Path) -> Tuple[Path, Path]:
    """
    Downloads and extracts KOfam data

    Args:
        outdir (Path): Directory of database to store data.

    Returns:
        Tuple[Path, Path]: Location of processed data
    """
    logger.info("Downloading KOfam data")
    # Set download paths
    db_dir = outdir / 'kofam_data'
    db_dir.mkdir(parents=True, exist_ok=True)
    profile_info_path = db_dir / 'ko_list.gz'
    ko_profiles_path = db_dir / 'profiles.tar.gz'
    profile_info_path_extr = db_dir / 'ko_list'
    ko_profiles_path_extr = db_dir / 'profiles'
    # Download latest annotations
    try:
        wget.download("ftp://ftp.genome.jp/pub/db/kofam/ko_list.gz",
                            out=str(profile_info_path))
        subprocess.call(['gunzip', str(profile_info_path)])
    except Exception:
        logger.warning(
            f"Data not available from kofam source. "
            f"Using backup data from XXX")
    # Download latest profiles
    try:
        wget.download("ftp://ftp.genome.jp/pub/db/kofam/profiles.tar.gz",
                            out=str(ko_profiles_path))
        logger.info("Extracting profiles")
        subprocess.call(['tar', 'xfz', str(ko_profiles_path),
        '-C', str(db_dir)])
        ko_profiles_path.unlink()
    except Exception:
        logger.warning(
            f"Data not available from kofam source. "
            f"Using backup data from XXX")
    
    # Write date accessed
    today = date.today()
    download_date = outdir / 'kofam_data/date_accessed.txt'
    with open(download_date, 'w') as output:
        output.write(f'{today.strftime("%B %d, %Y")}')
    logger.info("Finished")

    return profile_info_path_extr, ko_profiles_path_extr

# Format hmm models into groups
def kofam_formatter(ko_profiles_path_extr: Path) -> None:
    """
    Formats data downloaded for kofamscan

    Args:
        ko_profiles_path_extr (Path): Path to location of hmm profiles.
    """
    logger.info("Formatting KOfam profiles")  
    prokaryote_list = []
    eukaryote_list = []
    common_list = []
    with open(ko_profiles_path_extr / 'prokaryote.hal', 'r') as infile:
        for line in infile:
            prokaryote_list.append(line.strip())
    with open(ko_profiles_path_extr / 'eukaryote.hal', 'r') as infile:
        for line in infile:
            eukaryote_list.append(line.strip())
    # Get models shared by prokaryote and eukaryote
    for model, occurence in dict(Counter(prokaryote_list + eukaryote_list)).items():
        if occurence > 1:
            common_list.append(model)
    # Get unique modules per prokaryote or eukaryote
    prokaryote_only = list(set(prokaryote_list).difference(set(common_list)))
    eukaryote_only = list(set(eukaryote_list).difference(set(common_list)))
    # Get models not in a subset
    accounted_models = eukaryote_only + prokaryote_only + common_list
    total_models = []
    for file in ko_profiles_path_extr.iterdir():
        if file.suffix == '.hmm':
            total_models.append(file.name)
    orfan_models = list(set(total_models).difference(set(accounted_models)))
    # Write common models into groups of 500
    hmm_datasets = [common_list, prokaryote_only, eukaryote_only, orfan_models]
    hmm_datasets_names = ['common', 'prokaryote', 'eukaryote', 'independent']
    chunk_size = 500
    for hmm_list, name in zip(hmm_datasets, hmm_datasets_names):
        file_counter = 1
        file_list = []
        for i in range(0, len(hmm_list), chunk_size):
            group = hmm_list[i:i+chunk_size]
            with open(ko_profiles_path_extr/ f"{name}_{file_counter}.model", 'w') \
                as output:
                for file in group:
                    with open(ko_profiles_path_extr / file, 'r') as temporal:
                        shutil.copyfileobj(temporal, output)
                    (ko_profiles_path_extr / file).unlink()
            file_list.append(f"{name}_{file_counter}.model")
            file_counter += 1
        with open(ko_profiles_path_extr / f"{name}.list", 'w') as output:
            for file in file_list:
                output.write(f"{str(file)}\n")
    logger.info("Finished") 


# Parse kofamscan model info
def parse_profiles(profile_info_path: Path) -> Dict[str, namedtuple]:
    """
    Parses profile thresholds, score_type and definitions from kofamscan
    ko_list.

    Args:
        profile_info (Path): Path to ko_list file with profile information.

    Returns:
        Dict[str, namedtuple]: Dictionary with profile information.
    """
    logger.info("Parsing HMM profile metadata") 
    profile_data: Dict[str, namedtuple] = {}
    Profile = namedtuple('Profile', ["threshold", "score_type", "definition"])
    with open(profile_info_path, 'r') as infile:
        for line in infile:
            if line.startswith("knum"):
                continue
            else:
                line = line.strip().split("\t")
                if line[1] == '-':
                    profile_data[line[0]] = Profile(
                        line[1],line[2], line[11])
                else:
                    profile_data[line[0]] = Profile(
                        float(line[1]),line[2], line[11])

    return profile_data
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script downloads and parses KOfam HMM profiles\n"
            f"to be used by MicrobeAnnotator.\n"
            f"Mandatory parameters: -o [output directory]\n"))
    # Setup argument group
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-o', '--outdir', dest='outdir',action='store',
        required=True, help='Output directory.')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    outdir = arguments.outdir
    # Run functions
    profile_info_path_extr, ko_profiles_path_extr = kofam_downloader(outdir)
    kofam_formatter(ko_profiles_path_extr)
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================


#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Downloads the information necessary to create interconversion of
# accessions and appends it into the MicrobeAnnotator database.
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from pathlib import Path
from sys import argv

import argparse
import sqlite3
import wget
import gzip
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Create table with links from RefSeq -> UniProt
def create_refseq_to_uniprot(
    input_table: Path, sqlitedb: Path) -> None:
    """[summary]

    Args:
        input_table (Path): [description]
        database_to_append (Path): [description]
    """
    logger.info("Creating RefSeq to UniProt table")
    # Connect to database
    conn = sqlite3.connect(sqlitedb)
    cursor = conn.cursor()
    cursor.execute("DROP TABLE IF EXISTS refseq_to_uniprot")
    cursor.execute('CREATE TABLE refseq_to_uniprot \
        (refseq_id TEXT, uniprot_id TEXT)')
    # Read input table and create table within SQLite database
    with open(input_table, 'r') as refseq_conversion:
        record_counter = 0
        records = []
        for line in refseq_conversion:
            # Commit changes after 500000 records
            if record_counter == 500000:
                cursor.execute("begin")
                cursor.executemany(
                    'INSERT INTO refseq_to_uniprot VALUES(?, ?)', records)
                cursor.execute("commit")
                record_counter = 0
                records = []
            else:
                line = line.strip('\n').split("\t")
                records.append((line[1], line[0]))
                record_counter += 1
        # Commit remaining records
        if record_counter > 0:
            cursor.execute("begin")
            cursor.executemany(
                'INSERT INTO refseq_to_uniprot VALUES(?, ?)', records)
            cursor.execute("commit")
        # Create index for faster access
        cursor.execute(
            'CREATE INDEX refseq_index ON refseq_to_uniprot (refseq_id)')
    logger.info("Finished")

# Function to download and create KO to EC table 
def create_ko_to_ec(database_directory: Path, sqlitedb: Path) -> str:
    """
    Downloads and parses KO to EC identifier assignments and
    creates table in conversion.db

    Args:
        database_directory (Path): Directory for MicrobeAnnotator database.
        sqlitedb (Path): Path to SQLite database.

    Returns:
        str: Temporal KO_to_EC_identifiers.txt.gz file path.
    """
    logger.info("Creating KO to EC table")
    # Download information
    download_output = str(database_directory / "KO_to_EC_identifiers.txt.gz")
    down_url = (
        f"http://enve-omics.ce.gatech.edu/data/public_microbeannotator/"
        f"02.KO_to_EC_identifiers.txt.gz"
    )
    wget.download(down_url, out=download_output)
    # Connect with database
    conn = sqlite3.connect(sqlitedb)
    cursor = conn.cursor()
    # Create table with correspondence KO -> EC
    cursor.execute("DROP TABLE IF EXISTS ko_to_ec")
    cursor.execute('CREATE TABLE ko_to_ec \
        (ko_identifier TEXT, ec_identifier TEXT)')
    with gzip.open(download_output, 'rt') as ko_conversion:
        record_counter = 0
        records = []
        for line in ko_conversion:
            # Commit changes after 500000 records
            if record_counter == 50000:
                cursor.execute("begin")
                cursor.executemany('INSERT INTO ko_to_ec VALUES(?, ?)', records)
                cursor.execute("commit")
                record_counter = 0
                records = []
            else:
                line = line.strip('\n').split("\t")
                records.append((line[0], line[1]))
                record_counter += 1
        # Commit remaining records
        if record_counter > 0:
            cursor.execute("begin")
            cursor.executemany('INSERT INTO ko_to_ec VALUES(?, ?)', records)
            cursor.execute("commit")
        # Create index for faster access
    cursor.execute('CREATE INDEX ko_id ON ko_to_ec (ko_identifier)')
    cursor.execute('CREATE INDEX ec_id ON ko_to_ec (ec_identifier)')
    logger.info("Finished")

    return download_output

# Function to download and parse interpro tables
def create_interpro_tables(database_directory: Path, sqlitedb: Path) -> str:
    """
    Downloads and parses InterPro tables for intercoversion of ids.

    Args:
        database_directory (Path): Directory for MicrobeAnnotator database.
        sqlitedb (Path): Path to SQLite database.

    Returns:
        str: Path to temporal interpro_metadata.xml.gz file.
    """
    logger.info("Creating InterPro tables")
    # Download information
    download_output = str(database_directory / "interpro_metadata.xml.gz")
    wget.download("ftp://ftp.ebi.ac.uk/pub/databases/interpro/interpro.xml.gz",
                            out=download_output)
    interpro_to_ec = {}
    # Parse file information into dictionaries to merge into the SQLite database
    with gzip.open(download_output, 'rt') as infile:
        interpro_id = ""
        ec_identifier = []
        for line in infile:
            if line.startswith("<interpro id="):
                interpro_id = line.strip().split('"')[1]
            elif 'db_xref db="EC"' in line:
                identifier = line.strip().split('"')[3]
                ec_identifier.append(identifier)
            elif line.startswith("</interpro>"):
                if len(ec_identifier) > 0:
                    interpro_to_ec[interpro_id] = list(set(ec_identifier))
                interpro_id = ""
                ec_identifier = []

    # Connect with the database
    conn = sqlite3.connect(sqlitedb)
    cursor = conn.cursor()
    # Create table with correspondence InterPro -> EC
    cursor.execute("DROP TABLE IF EXISTS interpro_to_ec")
    cursor.execute('CREATE TABLE interpro_to_ec \
        (interpro_id TEXT, ec_identifier TEXT)')
    record_counter = 0
    records = []
    for interproscan, ec_id in interpro_to_ec.items():
        for ec_id_record in ec_id:
            # Commit changes after 5000 records
            if record_counter == 5000:
                cursor.execute("begin")
                cursor.executemany(
                    'INSERT INTO interpro_to_ec VALUES(?, ?)', records)
                cursor.execute("commit")
                record_counter = 0
                records = []
            else:
                records.append((interproscan, ec_id_record))
                record_counter += 1
    # Commit remaining records
    if record_counter > 0:
        cursor.execute("begin")
        cursor.executemany('INSERT INTO interpro_to_ec VALUES(?, ?)', records)
        cursor.execute("commit")
    # Create index for faster access
    cursor.execute(
        'CREATE INDEX interpro_index_ce ON interpro_to_ec (interpro_id)')
    logger.info("Finished")

    return download_output
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script downloads and parses tables required for\n"
            f"intercoversion of identifiers and stores in a SQLite database.\n"
            f"Mandatory parameters: -i [input refseq-uniprot table] \n"
            f"-d [database directory] -s [sqlite database]\n"))
    # Setup mandatory arguments
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-i', '--input_table', dest='intable', action='store', required=True,
        help='RefSeq to Uniprot table to create correspondence table.')
    mandatory_arguments.add_argument(
        '-d', '--database', dest='database', action='store', required=True,
        help='Database directory where data should be stored.')
    mandatory_arguments.add_argument(
        '-s', '--sqlitedb', dest='sqlitedb', action='store', required=True,
        help='SQLite database to create tables in.')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    intable = arguments.intable
    database = arguments.database
    sqlitedb = arguments.sqlitedb
    # Run functions
    try:
        create_refseq_to_uniprot(intable, sqlitedb)
    except:
        logger.error("Could not create RefSeq to Uniprot table.")
        exit(1)
    try:
        create_ko_to_ec(database, sqlitedb)
    except:
       logger.error("Could not create KO to EC table.")
       exit(1)
    try:
        create_interpro_tables(database, sqlitedb)
    except:
       logger.error("Could not create InterPro tables.")
       exit(1)
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================



#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz-Perez
# Email:        cruizperez3@gatech.edu
# Institution:  Georgia Institute of Technology
# Date:         11 April 2021

# Description: This script defines the logger and its options
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
import logging
import socket
# ==============================================================================


# ==============================================================================
# Define logging methods and filter classes
# ==============================================================================
# Class to add filters for handlers
class LoggerFilter(object):
    def __init__(self, level):
        self.__level = level
    
    def filter(self, logRecord):
        return logRecord.levelno <= self.__level

# Setup logger with handlers for different levels
def setup_logger(name, level: str = 'INFO'):
    # Define format for handlers
    info_format = (f"%(asctime)s " + "[%(levelname)s]: %(message)s")
    warning_format = (
        f"%(asctime)s {socket.gethostname()} %(name)s: "
        f"Function: %(funcName)s - Line: %(lineno)s "
        + "[%(levelname)s]: %(message)s")
    # Define handler for information
    info_handler = logging.StreamHandler()
    info_handler.setLevel(logging.INFO)
    info_handler.setFormatter(logging.Formatter(info_format))
    info_handler.addFilter(LoggerFilter(logging.INFO))
    # Define handler for warnings
    warning_handler = logging.StreamHandler()
    warning_handler.setLevel(logging.WARNING)
    warning_handler.setFormatter(logging.Formatter(warning_format))
    warning_handler.addFilter(LoggerFilter(logging.WARNING))
    # Define handler for errors and critical
    error_handler = logging.StreamHandler()
    error_handler.setLevel(logging.ERROR)
    error_handler.setFormatter(logging.Formatter(warning_format))
    # Initialize logger
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    logger.addHandler(info_handler)
    logger.addHandler(warning_handler)
    logger.addHandler(error_handler)

    return logger
# ==============================================================================


#!/usr/bin/env python

"""
########################################################################
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Maps protein KO information with their respective modules
# and calculates the completeness percentage of each module present.
########################################################################
"""

################################################################################
"""---0.0 Import Modules---"""
import pandas as pd
import re
import pickle
from pathlib import Path
import matplotlib.pyplot as plt
import seaborn as sns
from collections import OrderedDict
from sys import exit
import matplotlib

################################################################################
"""---1.0 Define Functions---"""

def ko_match(string):
    """ Looks if string has the form K[0-9]{5}
    
    Arguments:
        string {string} -- String to test
    
    Returns:
        [bool] -- String has form or not
    """
    if re.search(r'^K[0-9]{5}$', string) is not None:
        return 1
    else:
        return 0

def split_compound(string, comp_type):
    """[summary]
    
    Arguments:
        string {[type]} -- [description]
        comp_type {[type]} -- [description]
    
    Returns:
        [type] -- [description]
    """
    if comp_type == 'compound':
        return re.split('[-+]', string)
    elif comp_type == 'and_comp':
        return string.split('_')
    elif comp_type == 'or_comp':
        return string.split(',')

def process_compounds(string, comp_type, genome_annotation):
    string = split_compound(string, comp_type)
    proteins_required = len(string)
    proteins_present = 0
    for option in string:
        if option == '':
            proteins_required -= 1
        elif '+' in option or '-' in option:
            compound = split_compound(option, 'compound')
            proteins_in_compound = len(compound)
            present_in_compound = 0
            for sub_option in compound:
                if ko_match(sub_option) > 0 and sub_option in genome_annotation:
                    present_in_compound += 1
            proteins_present += present_in_compound/proteins_in_compound
        else:
            if ko_match(option) > 0 and option in genome_annotation:
                proteins_present += 1
    return proteins_present, proteins_required

def import_module_dict(dictionary_location):
    pickle_in = open(dictionary_location,"rb")
    dictionary = pickle.load(pickle_in)
    return dictionary

def module_information_importer(input_files):
    print("Importing data... ", end="")
    script_path = Path(__file__)
    script_dir = script_path.parent
    data_folder = Path(script_dir).parent / "data"
    # Import all modules from dictionaries
    regular_modules = import_module_dict(data_folder / "01.KEGG_Regular_Module_Information.pickle")
    bifurcation_modules = import_module_dict(data_folder / "02.KEGG_Bifurcating_Module_Information.pickle")
    structural_modules = import_module_dict(data_folder / "03.KEGG_Structural_Module_Information.pickle")
    # Get genome and module ids
    module_information = {}
    module_ids = []
    genome_names = []
    group_names = []
    # Fill genome names list
    for genome_file in input_files:
        genome_names.append(str(Path(genome_file).name))
    # Get module correspondence
    with open(data_folder / "01.KEGG_DB/00.Module_Names.txt") as correspondence:
        for line in correspondence:
            line = line.strip().split("\t")
            module_information[line[0]] = [line[1], line[2], line[3]]
            # module_id_name[line[0]] = line[1]
            module_ids.append(line[0])
            # module_groups[line[0]] = line[2]
            if line[2] not in group_names:
                group_names.append(line[2])
            # module_colors[line[1]] = [line[2],line[3]]
    # Create module pathway group matrix
    module_group_matrix = pd.DataFrame(0, index=group_names, columns=genome_names)
    # Create module matrix
    metabolism_matrix = pd.DataFrame(0, index=module_ids, columns=genome_names)
    print("Finished")
    return regular_modules, bifurcation_modules, structural_modules, \
           module_information, metabolism_matrix, module_group_matrix

def annotation_file_parser(genome_annotation_file):
    genome_annotation = []
    with open(genome_annotation_file) as annot_input:
        for line in annot_input:
            genome_annotation.append(line.strip())
    return genome_annotation

def regular_module_mapper(module_dictionary, genome_annotation_file):
    genome_annotation = annotation_file_parser(genome_annotation_file)
    regular_module_completenes = []
    for module, final_steps in module_dictionary.items():
        complete_steps = 0
        total_module_steps = len(final_steps)
        for proteins in final_steps.values():
            score_for_step = 0
            steps_per_option = 100
            match = False
            for option in proteins:
                match = False
                # Search for single gene option
                if ko_match(option) > 0:
                    if option in genome_annotation:
                        score_for_step = 1
                        match = True
                    if match == True:
                        steps_per_option = 1
                    elif 1 < steps_per_option:
                        steps_per_option = 1
                elif '%' in option:
                    option = option.replace(')', '')
                    option = option.split('-%')
                    mandatory = split_compound(option[0], 'compound')
                    proteins_required = len(mandatory) + 1
                    proteins_present = 0
                    for prot in mandatory:
                        if ko_match(prot) > 0 and prot in genome_annotation:
                            proteins_present += 1
                    for prot in option[1].split(','):
                        if ko_match(prot) > 0 and prot in genome_annotation:
                            proteins_present += 1
                            break
                    if proteins_present/proteins_required > score_for_step:
                        score_for_step = proteins_present/proteins_required
                        match = True
                    if match == True:
                        steps_per_option = 1
                    elif 1 < steps_per_option:
                        steps_per_option = 1
                elif '_' in option and ',' in option:
                    option = sorted(split_compound(option, 'and_comp'), key=len)
                    highest_score = 0
                    for element in option:
                        if ko_match(element) > 0 and element in genome_annotation:
                            highest_score += 1
                        elif ',' in element:
                            element = sorted(element.split(","), key=len)
                            for sub_element in element:
                                if ko_match(sub_element) > 0 and sub_element in genome_annotation:
                                    highest_score += 1
                                    break
                                elif '+' in sub_element:
                                    proteins_present, proteins_required = process_compounds(sub_element, 
                                    'compound', genome_annotation)
                                    highest_score += proteins_present/proteins_required
                    if highest_score > score_for_step:
                        score_for_step = highest_score
                        match = True
                    if match == True:
                        steps_per_option = len(option)
                    elif len(option) < steps_per_option:
                        steps_per_option = len(option)
                elif len(option.split(",")) > 1:
                    match = False
                    option = sorted(option.split(","), key=len)
                    highest_score = 0
                    steps_to_add = 0
                    for sub_option in option:
                        if ko_match(sub_option) > 0 and sub_option in genome_annotation:
                            if 1 > highest_score:    
                                highest_score = 1
                                steps_to_add = 1
                        elif '+' in sub_option or '-' in sub_option:
                            proteins_present, proteins_required = process_compounds(sub_option, 'compound', genome_annotation)
                            if proteins_present/proteins_required > highest_score:
                                highest_score = proteins_present/proteins_required
                                steps_to_add = 1
                    if highest_score > score_for_step:
                        score_for_step = highest_score
                        match = True
                    if match == True:
                        steps_per_option = steps_to_add
                    elif steps_to_add< steps_per_option:
                        steps_per_option = steps_to_add
                elif '_' in option:
                    match = False
                    proteins_present, proteins_required = process_compounds(option, 'and_comp', genome_annotation)
                    if proteins_present/proteins_required > score_for_step:
                        score_for_step = proteins_present
                        match = True
                    if match == True:
                        steps_per_option = proteins_required
                    elif proteins_required < steps_per_option:
                        steps_per_option = proteins_required
                elif "+" in option or "-" in option:
                    match = False
                    highest_score = 0
                    proteins_present, proteins_required = process_compounds(option, 'compound', genome_annotation)
                    if proteins_present/proteins_required > score_for_step:
                        score_for_step = proteins_present/proteins_required
                        match = True
                    if match == True:
                        steps_per_option = 1
                    elif proteins_required < steps_per_option:
                        steps_per_option = 1
                else:
                    print("Unreccognized module {}. Check your database.".format(option))
            complete_steps += score_for_step
            if steps_per_option > 50:
                steps_per_option = 1
            if steps_per_option > 1:
                total_module_steps += steps_per_option - 1
        regular_module_completenes.append((module, round((complete_steps/total_module_steps)*100, 2)))
    return regular_module_completenes

def bifurcating_module_mapper(module_dictionary, genome_annotation_file):
    genome_annotation = annotation_file_parser(genome_annotation_file)
    bifurcating_module_completenes = []
    for module, versions in module_dictionary.items():
        module_highest = 0
        for version, total_steps in versions.items():
            completed_steps = 0
            total_version_steps = len(total_steps)
            for proteins in total_steps.values():
                score_for_step = 0
                steps_per_option = 100
                match = False
                if isinstance(proteins, (list)):
                    protein = sorted(proteins, key=len)
                    for option in protein:
                        match = False
                        if ko_match(option) > 0:
                            if option in genome_annotation:
                                score_for_step = 1
                                match = True
                            if match == True:
                                steps_per_option = 1
                            elif 1 < steps_per_option:
                                steps_per_option = 1
                        elif '_' in option and ',' in option:
                            option = sorted(split_compound(option, 'and_comp'), key=len)
                            highest_score = 0
                            for element in option:
                                if ko_match(element) > 0 and element in genome_annotation:
                                    highest_score += 1
                                elif ',' in element:
                                    element = sorted(element.split(","), key=len)
                                    score_sub_option = 0
                                    for sub_element in element:
                                        if ko_match(sub_element) > 0 and sub_element in genome_annotation:
                                            highest_score += 1
                                            break
                                        elif '+' in sub_element:
                                            proteins_present, proteins_required = process_compounds(sub_element, 
                                            'compound', genome_annotation)
                                            highest_score += proteins_present/proteins_required
                            if highest_score > score_for_step:
                                score_for_step = highest_score
                                match = True
                            if match == True:
                                steps_per_option = len(option)
                        elif '_' in option:
                            match = False
                            proteins_present, proteins_required = process_compounds(option, 'and_comp', genome_annotation)
                            if proteins_present/proteins_required > score_for_step:
                                score_for_step = proteins_present
                                match = True
                            if match == True:
                                steps_per_option = proteins_required
                            elif proteins_required < steps_per_option:
                                steps_per_option = proteins_required
                        elif '+' in option or '-' in option:
                            match = False
                            highest_score = 0
                            proteins_present, proteins_required = process_compounds(option, 'compound', genome_annotation)
                            if proteins_present/proteins_required > score_for_step:
                                score_for_step = proteins_present/proteins_required
                                match = True
                            if match == True:
                                steps_per_option = 1
                            elif proteins_required < steps_per_option:
                                steps_per_option = 1
                elif ko_match(proteins) > 0:
                    match = False
                    if proteins in genome_annotation:
                        score_for_step = 1
                        match = True
                    if match == True:
                        steps_per_option = 1
                    elif 1 < steps_per_option:
                        steps_per_option = 1
                elif ',' in proteins:
                    options = split_compound(proteins, 'or_comp')
                    for option in options:
                        if ko_match(option) > 0 and option in genome_annotation:
                            score_for_step = 1
                            match = True
                        if match == True:
                            steps_per_option = 1
                        elif 1 < steps_per_option:
                            steps_per_option = 1
                elif '+' in proteins or '-' in proteins:
                    match = False
                    highest_score = 0
                    proteins_present, proteins_required = process_compounds(proteins, 'compound', genome_annotation)
                    if proteins_present/proteins_required > score_for_step:
                        score_for_step = proteins_present/proteins_required
                        match = True
                    if match == True:
                        steps_per_option = 1
                    elif proteins_required < steps_per_option:
                        steps_per_option = 1
                else:
                    print("Unreccognized module {}. Check your database.".format(proteins))
                completed_steps += score_for_step
                if steps_per_option > 50:
                    steps_per_option = 1
                if steps_per_option > 1:
                    total_version_steps += steps_per_option - 1
            if completed_steps/total_version_steps > module_highest:
                module_highest = completed_steps/total_version_steps
        bifurcating_module_completenes.append((module, round(module_highest*100, 2)))
    return bifurcating_module_completenes

def structural_module_mapper(module_dictionary, genome_annotation_file):
    genome_annotation = annotation_file_parser(genome_annotation_file)
    structural_module_completenes = []
    for module, components in module_dictionary.items():
        score_for_components = 0
        module_proteins_present = 0
        module_proteins_required = 0
        for proteins in components:
            if isinstance(proteins, (list)):
                highest_score = 0
                proteins_to_add = 0
                steps_to_add = 100
                for option in proteins:
                    if '_' in option and ',' in option:
                        option = sorted(split_compound(option, 'and_comp'), key=len)
                        proteins_present_option = 0
                        proteins_required_option = 0
                        for element in option:
                            if ko_match(element) > 0:
                                if element in genome_annotation:
                                    proteins_present_option += 1
                                    proteins_required_option += 1
                                else:
                                    proteins_required_option += 1
                            elif ',' in element:
                                element = sorted(element.split(","), key=len)
                                score_sub_element = 0
                                proteins_present_sub_element = 0
                                proteins_required_sub_element = 0
                                for sub_element in element:
                                    if ko_match(sub_element) > 0:
                                        if sub_element in genome_annotation:
                                            score_sub_element = 1
                                            proteins_present_sub_element += 1
                                            proteins_required_sub_element += 1
                                        elif 1 < proteins_required_sub_element and score_sub_element == 0:
                                            proteins_required_sub_element = 1
                                    elif '+' in sub_element or '-' in sub_element:
                                        proteins_present, proteins_required = process_compounds(sub_element, 
                                        'compound', genome_annotation)
                                        if proteins_present/proteins_required > score_sub_element:
                                            score_sub_element = proteins_present/proteins_required
                                            proteins_present_sub_element = proteins_present
                                            proteins_required_sub_element = proteins_required
                                        elif proteins_required < proteins_required_sub_element and score_sub_element == 0:
                                            proteins_required_sub_element = proteins_required
                                proteins_present_option += proteins_present_sub_element
                                proteins_required_option += proteins_required_sub_element
                        if proteins_present_option/proteins_required_option > highest_score:
                            highest_score = proteins_present_option/proteins_required_option
                            proteins_to_add = proteins_present_option
                            steps_to_add = proteins_required_option
                        elif proteins_required_option < steps_to_add and highest_score == 0:
                            steps_to_add = proteins_required_option
                    elif '+' in option or '-' in option:
                        proteins_present, proteins_required = process_compounds(option, 'compound', genome_annotation)
                        if proteins_present/proteins_required > highest_score:
                            highest_score = proteins_present/proteins_required
                            steps_to_add = proteins_required
                            proteins_to_add = proteins_present
                        elif proteins_required < steps_to_add and highest_score == 0:
                            steps_to_add = proteins_required
                module_proteins_present += proteins_to_add
                module_proteins_required += steps_to_add
            elif ko_match(proteins) > 0:
                if proteins in genome_annotation:
                        module_proteins_present += 1
                        module_proteins_required += 1
                else:
                    module_proteins_required += 1
            elif ',' in proteins:
                proteins = sorted(proteins.split(","), key=len)
                score = 0
                proteins_present_option = 0
                proteins_required_option = 100
                for element in proteins:
                    if ko_match(element) > 0:
                        if element in genome_annotation:
                            proteins_required_option = 1
                            proteins_present_option = 1
                            break
                        elif 1 < proteins_required_option and score == 0:
                            proteins_required_option = 1
                    elif '+' in element or '-' in element:
                        proteins_present, proteins_required = process_compounds(element, 'compound', genome_annotation)
                        if proteins_present/proteins_required > score:
                            score = proteins_present/proteins_required
                            proteins_present_option = proteins_present
                            proteins_required_option = proteins_required
                        elif proteins_required < proteins_required_option and score == 0:
                            proteins_required_option = proteins_required
                module_proteins_present += proteins_present_option
                module_proteins_required += proteins_required_option
            elif '+' in proteins or '-' in proteins:
                proteins_present, proteins_required = process_compounds(proteins, 'compound', genome_annotation)
                module_proteins_present += proteins_present
                module_proteins_required += proteins_required  
            else:
                print("Unreccognized module {}. Check your database.".format(proteins))
        score_for_components = module_proteins_present/module_proteins_required
        structural_module_completenes.append((module, round(score_for_components*100, 2)))
    return structural_module_completenes

def global_mapper(regular_modules, bifurcating_modules, structural_modules, annotation_files):
    print("Mapping entries... ", end="")
    full_metabolic_completeness = {}
    
    for file in annotation_files:
        genome_name = str(Path(file).name)
        regular_completeness = regular_module_mapper(regular_modules, file)
        bifurcating_completeness = bifurcating_module_mapper(bifurcating_modules, file)
        structural_completeness = structural_module_mapper(structural_modules, file)
        final_completeness = regular_completeness + bifurcating_completeness + structural_completeness
        full_metabolic_completeness[genome_name] = final_completeness
    print("Finished\n")
    return full_metabolic_completeness

def plot_function_barplots(module_colors, module_group_matrix, metabolism_matrix_dropped_relabel, prefix):
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.use('Agg')
    print("Grouping by metabolism type and plotting... ")
    for module in list(metabolism_matrix_dropped_relabel.index):
        for genome in list(metabolism_matrix_dropped_relabel.columns):
            if metabolism_matrix_dropped_relabel.loc[module,genome] >= 80:
                module_group_matrix.loc[module_colors[module][0],genome] += 1
    module_group_matrix_transp = module_group_matrix.T
    emptyness = (module_group_matrix_transp == 0).all()
    if emptyness.all() == True:
        print("There are no modules above 80% completeness. No barplot will be generated.")
    else:
        module_group_matrix_transp = module_group_matrix_transp.loc[(module_group_matrix_transp >= 1).any(1)]
        color_dict = {}
        color_list = []
        for pair in module_colors.values():
            if pair[0] not in color_dict:
                color_dict[pair[0]] = pair[1]
        for pathway in list(module_group_matrix_transp.columns):
            color_list.append(color_dict[pathway])
        Figure, Axis = plt.subplots(1,1, figsize=(25,15))
        module_group_matrix_transp.plot.bar(ax=Axis, stacked=True, color=color_list, figsize=(25,15), legend=False)
        Axis.legend(loc="center left", bbox_to_anchor=(1.02, 0.5), fontsize='medium', markerscale=0.3)
        Axis.set_xlabel("Genomes", fontsize=30)
        Axis.set_ylabel('Number of Modules (>=80% complete)', fontsize=30)
        Axis.tick_params(axis='x', labelsize=15)
        Axis.tick_params(axis='y', labelsize=15)
        Figure.suptitle('Metabolism Module Category per Genome', fontsize=40)
        Figure.subplots_adjust()
        Figure.savefig(prefix + "_barplot.pdf", bbox_inches="tight")
    print("Finished")

def create_output_files(metabolic_annotation, metabolism_matrix, module_information, cluster, prefix):
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.use('Agg')
    print("Creating output_file matrix and heatmap... ")
    # Check clustering asked
    if cluster is None:
        row_cluster = False
        col_cluster = False
    elif cluster == 'rows':
        row_cluster = True
        col_cluster = False
    elif cluster == 'cols':
        row_cluster = False
        col_cluster = True
    elif cluster == 'both':
        row_cluster = True
        col_cluster = True
    else:
        exit("Wrong clustering option, select between 'both', 'cols', 'rows' or don't pass this option.")
    # Extract data from metabolism matrix and color dictionary
    for genome, modules in metabolic_annotation.items():
        # Fill metabolism matrix with completeness from the metabolic annotation dictionary
        for module in modules:
            metabolism_matrix.loc[module[0],genome] = module[1]
    # Get module id and its corresponding module name and the module and its corresponding color
    module_id_name = {}
    module_colors = {}
    for module, information in module_information.items():
        module_id_name[module] = information[0]
        module_colors[information[0]] = (information[1],information[2])
    # Get modules that are above 50% complete in at least one genome
    ylabel_text = 'Modules (at least 50% complete in at least one genome)'
    metabolism_matrix_retained = metabolism_matrix.loc[(metabolism_matrix >= 50).any(1)]
    table_empty = (metabolism_matrix_retained == 0).all()
    if table_empty.all() == True:
        print("There are no modules above 50% completeness. Plotting all modules regardless of completeness.")
        metabolism_matrix_retained = metabolism_matrix.loc[(metabolism_matrix > 0).any(1)]
        ylabel_text = 'Modules'
    colors_for_ticks = []
    colors_for_legend = {}
    metabolism_matrix_retained_relabel = metabolism_matrix_retained.rename(index=module_id_name)
    # Save original annotation table but add module names and pathway
    module_names = []
    module_pathways = []
    for module in metabolism_matrix.index:
        module_names.append(module_information[module][0])
        module_pathways.append(module_information[module][1])
    metabolism_matrix.insert(loc=0, column="name", value=module_names)
    metabolism_matrix.insert(loc=1, column="pathway group", value=module_pathways)
    metabolism_matrix.index.name = 'module'
    metabolism_matrix.to_csv(prefix + "_module_completeness.tab", sep="\t")
    
    # Plot heatmap with clustering
    # Check if clustering is possible with the number of genomes and modules
    table_dim = metabolism_matrix_retained_relabel.shape
    if row_cluster == True and table_dim[0] < 3:
        row_cluster = False
    if col_cluster == True and table_dim[1] < 3:
        col_cluster = False
    sns.set(font_scale=1.6)
    heatmap = sns.clustermap(metabolism_matrix_retained_relabel, xticklabels=True, yticklabels=True, 
                            figsize=(30,25), row_cluster=row_cluster, col_cluster=col_cluster, cbar_pos=(0.01, 1, 0.2, 0.04),
                            cbar_kws= {'orientation': 'horizontal', 'label': 'Module Completeness (%)'}, dendrogram_ratio=0.1)
    # cbar = heatmap.ax_cbar
# here set the labelsize by 20
    # cbar.set_yticklabels(labels=cbar.get_ylabel, fontdict={"fontsize":20})
    for module in heatmap.ax_heatmap.yaxis.get_ticklabels():
        colors_for_ticks.append(module_colors[module.get_text()][1])
        if module_colors[module.get_text()][0] not in colors_for_legend:
            colors_for_legend[module_colors[module.get_text()][0]] = module_colors[module.get_text()][1]
    [mod.set_color(col) for (col,mod) in zip(colors_for_ticks,heatmap.ax_heatmap.yaxis.get_ticklabels())]
    for pathway, color in colors_for_legend.items():
        heatmap.ax_heatmap.scatter(1,1, color=color, label=pathway, zorder=0, s=45)
    heatmap.ax_heatmap.set_xticklabels(heatmap.ax_heatmap.get_xmajorticklabels(), fontsize = 15)
    heatmap.ax_heatmap.set_yticklabels(heatmap.ax_heatmap.get_ymajorticklabels(), fontsize = 15)
    heatmap.ax_heatmap.set_xlabel("Genomes", fontsize=20)
    heatmap.ax_heatmap.set_ylabel(ylabel_text, fontsize=20)
    plt.figlegend(loc="upper right", ncol=2, fontsize=12)
    heatmap.savefig(prefix + "_heatmap.pdf")
    print("Finished")
    return metabolism_matrix_retained_relabel, module_colors

################################################################################
"""---2.0 Main Function---"""

def main():
    import argparse, sys
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
            description='''This maps KEGG KO identifiers into modules and summarizes the metabolic potential\n'''
            '''Usage: ''' + sys.argv[0] + ''' -i [Input Files] -p [Prefix Output]\n'''
            '''Global mandatory parameters: -i [Input Files] -p [Prefix Output]\n'''
            '''Optional Database Parameters: See ''' + sys.argv[0] + ' -h')
    parser.add_argument('-i', '--input_files', dest='input_files', nargs='+', action='store', required=True,
                        help='Space-separated list of files to parse.')
    parser.add_argument('-p', '--prefix', dest='prefix', action='store', required=True,
                        help='Prefix for the output_file files.')
    parser.add_argument('--cluster', dest='cluster', action='store', required=False,
                        help='Cluster genomes or modules. Select "cols" for genomes, "rows" for modules, or "both".\
                            By default no clustering.')
    args = parser.parse_args()

    input_files = args.input_files
    prefix = args.prefix
    cluster = args.cluster

    # ----------------------------
    # Import data
    regular_modules, bifurcation_modules, structural_modules,  \
    module_information, metabolism_matrix, module_group_matrix = module_information_importer(input_files)
    # Map and annotate genomes
    metabolic_annotation = global_mapper(regular_modules, bifurcation_modules, structural_modules, input_files)
    metabolism_matrix_dropped_relabel, module_colors = create_output_files(metabolic_annotation, metabolism_matrix, module_information, cluster, prefix)
    plot_function_barplots(module_colors, module_group_matrix, metabolism_matrix_dropped_relabel, prefix)
    # ----------------------------

if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
########################################################################
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Iteratively annotates proteins using different databases
# and multiple search methods.
########################################################################
"""

################################################################################
"""---0.0 Import Modules---"""
import wget
from shutil import copyfileobj
from pathlib import Path
import urllib.request
import gzip
from shutil import which
from shutil import rmtree

################################################################################
"""---1.0 Define Functions---"""
def kofamscan_annotation(protein_file, multi_argument):
    # Import modules
    import subprocess
    import random
    import string
    # Unpack arguments
    output_directory = multi_argument[0]
    threads = multi_argument[1]
    bin_path = multi_argument[2]
    # Create folders
    output_directory = Path(output_directory)
    kofamscan_result_folder = output_directory / 'kofamscan_results'
    kofamscan_result_folder.mkdir(parents=True, exist_ok=True)
    protein_file_name = Path(protein_file).name
    # Call kofamscan and filter results
    if bin_path == None:
        kofam_call = 'exec_annotation'
    else:
        kofam_call = str(Path(bin_path) / 'exec_annotation')
    kofam_output_file = str(kofamscan_result_folder / (protein_file_name + '.kofam'))
    temp_folder = ''.join([random.choice(string.ascii_letters + string.digits) for n in range(15)])
    subprocess.call([kofam_call, '-o', kofam_output_file, '--cpu', str(threads),
                    '--tmp-dir', temp_folder, protein_file])
    filtered_output_file = kofam_output_file + '.filt'
    kofamscan_filter(kofam_output_file, filtered_output_file)
    if Path(temp_folder).is_dir():
        rmtree(temp_folder)

    # Extract ids of proteins annotated and add annotations into final file
    final_annotation_folder = output_directory / 'annotation_results'
    final_annotation_folder.mkdir(parents=True, exist_ok=True)
    final_file = Path(final_annotation_folder) / (protein_file_name + '.annotations')
    ids_proteins_annotated = []
    ids_hypothetical_proteins = []
    #* The final annotation will have the following columns
    #* query_id protein_id product ko_number ko_product taxonomy function compartment process EC InterPro Pfam database
    with open(filtered_output_file, 'r') as ko_annotated, open(final_file, 'w') as output:
        output.write("query_id\tprotein_id\tproduct\tko_number\tko_product\ttaxonomy\tfunction_go\tcompartment_go\tprocess_go\tinterpro\tpfam\tec_number\tdatabase\n")
        for line in ko_annotated:
            if "#" in line:
                continue
            else:
                line = line.strip().split("\t")
                if "hypothetical" not in line[6].lower() and \
                    "uncharacterized" not in line[6].lower() and \
                    "unknown function" not in line[6].lower():
                    ids_proteins_annotated.append(line[1])
                else:
                    ids_hypothetical_proteins.append(line[1])
                output.write("{}\tNA\tNA\t{}\t{}\tNA\tNA\tNA\tNA\tNA\tNA\tNA\tkofamscan\n".format(line[1], line[2], line[6]))
    return (protein_file, protein_file_name, ids_proteins_annotated, final_file, ids_hypothetical_proteins)


def kofamscan_filter(kofamscan_input, outfile):
    # Import modules
    from random import randint
    from sys import exit

    header_list = []
    kofamscan_results = {}
    with open(kofamscan_input, 'r') as input:
        for line in input:
            if line.startswith("#"):
                if 'gene name' in line:
                    headers = []
                    headers.append('')
                    head = line.strip().split()
                    name = ' '.join(head[0:3])
                    headers.append(name)
                    headers += head[3:]
                    header_list.append("\t".join(headers))
                else:
                    header = ['']
                    header += line.strip().split()
                    header_list.append("\t".join(header))
            elif line.startswith('*'):
                result = line.strip().split()
                gene_name = result[1]
                annot_desc = ' '.join(result[6:])
                if gene_name not in kofamscan_results:
                    kofamscan_results[gene_name] = result[0:6]
                    kofamscan_results[gene_name].append(annot_desc)
                else:
                    if result[4] > kofamscan_results[gene_name][4]:
                        kofamscan_results[gene_name] = result[0:6]
                        kofamscan_results[gene_name].append(annot_desc)
                    elif result[4] == kofamscan_results[gene_name][4]:
                        if randint(0,1) > 0:
                            kofamscan_results[gene_name] = result[0:6]
                            kofamscan_results[gene_name].append(annot_desc)
                        else:
                            continue
                    else:
                        continue
            else:
                continue

    with open(outfile, 'w') as output_file:
        for element in header_list:
            output_file.write("{}\n".format(element))
        for hit in kofamscan_results.values():
            output_file.write("{}\n".format("\t".join(hit)))


def blast_filter_slow(input_tab, outfile, id_perc, bitscore, evalue, aln_percent):
    # Import modules
    from random import choice

    blast_hits = {}
    # Retrieve best matches
    with open(input_tab) as blast_input:
        for line in blast_input:
            line = line.strip()
            hit = line.split("\t")
            if float(hit[2]) >= id_perc and float(hit[11]) >= bitscore \
            and float(hit[10]) <= evalue and float(hit[3])*100/float(hit[12]) >= aln_percent:
                if hit[0] not in blast_hits:
                    blast_hits[hit[0]] = [float(hit[11]), [line]]
                else:
                    if float(hit[11]) < blast_hits[hit[0]][0]:
                        continue
                    elif float(hit[11]) > blast_hits[hit[0]][0]:
                        blast_hits[hit[0]] = [float(hit[11]), [line]]
                    else:
                        blast_hits[hit[0]][1].append(line)
            else:
                continue
    with open(outfile, 'w') as output:
        for hit_values in blast_hits.values():
            output.write("{}\n".format(choice(hit_values[1])))


def add_query_len(input_tab, protein_file, outfile):
    # Import modules
    from Bio.SeqIO.FastaIO import SimpleFastaParser

    # Read protein lenghts
    protein_len = {}
    with open(protein_file, 'r') as proteins:
        for title, seq in SimpleFastaParser(proteins):
            protein_len[title.split()[0]] = len(seq)
    with open(input_tab, 'r') as blast_like, open(outfile, 'w') as output:
        for line in blast_like:
            line = line.strip()
            protein = line.split("\t")[0]
            output.write("{}\t{}\n".format(line, protein_len[protein]))

def similarity_search(protein_file, multiple_arguments):
    # Import modules
    import subprocess
    from shutil import which
    from sys import exit
    # Unpack arguments
    output_directory = multiple_arguments[0]
    db_version = multiple_arguments[1]
    database = multiple_arguments[2]
    method = multiple_arguments[3]
    threads = multiple_arguments[4]
    id_perc = multiple_arguments[5]
    bitscore = multiple_arguments[6]
    evalue = multiple_arguments[7]
    aln_percent = multiple_arguments[8]
    bin_path = multiple_arguments[9]
    # Create folders
    output_directory = Path(output_directory)
    db_result_folder = output_directory / (db_version + '_results')
    db_result_folder.mkdir(parents=True, exist_ok=True)
    protein_file_name = Path(protein_file).name

    # Call method and filter results
    if method == "blast":
        if bin_path == None:
            blast_call = "blastp"
        else:
            blast_call = str(Path(bin_path) / "blastp")
        if which(blast_call) == None:
            if bin_path == None:
                exit("Blastp not found in PATH, please provide the correct path to the folder with binaries")
            else:
                exit("Blastp not found in " + bin_path + ", please provide the correct path to the folder with binaries")
        else:
            method_output_file = str(db_result_folder / (protein_file_name + '.' + method))
            subprocess.call([blast_call, "-query", protein_file, "-db", database, "-out", method_output_file,
            "-max_target_seqs", "6", "-num_threads", str(threads), "-outfmt", """6 std qlen slen"""])
            filtered_file = method_output_file + '.filt'
            blast_filter_slow(method_output_file, filtered_file, id_perc, bitscore, evalue, aln_percent)
            return (protein_file, filtered_file)
    elif method == 'diamond':
        if bin_path == None:
            diamond_call = "diamond"
        else:
            diamond_call = str(Path(bin_path) / "diamond")
        # if which(method) == None:
        #     if bin_path == None:
        #         exit("Diamond not found in PATH, please provide the correct path to the folder with binaries")
        #     else:
        #         exit("Diamond not found in " + bin_path + ", please provide the correct path to the folder with binaries")
        # else:
        method_output_file = str(db_result_folder / (protein_file_name + '.' + method))
        subprocess.call([diamond_call, 'blastp', '--db', database, '--out', method_output_file, 
        '--outfmt', "6", "qseqid", "sseqid", "pident", "length", "mismatch", "gapopen", "qstart", "qend",
        "sstart", "send", "evalue", "bitscore", "qlen", "slen", "--threads", str(threads),
        "--unal", "0", "-b", "6", "--quiet", "--max-target-seqs", "6", "--query", protein_file])
        filtered_file = method_output_file + '.filt'
        blast_filter_slow(method_output_file, filtered_file, id_perc, bitscore, evalue, aln_percent)
        return (protein_file, filtered_file)
    elif method == 'sword':
        if bin_path == None:
            sword_call = "sword"
        else:
            sword_call = str(Path(bin_path) / "sword")
        if which(sword_call) == None:
            if bin_path == None:
                exit("Sword not found in PATH, please provide the correct path to the folder with binaries")
            else:
                exit("Sword not found in " + bin_path + ", please provide the correct path to the folder with binaries")
        else:
            method_output_file = str(db_result_folder / (protein_file_name + '.' + method))
            subprocess.call([sword_call, '-i', protein_file, '-j', database, 
            '-o', method_output_file, "-f", "bm8", "-a", "10000", "-k", "5", "-c", "15000",
            "-T", "16", "-t", str(threads)])
            temporal_file = method_output_file + '.temp'
            add_query_len(method_output_file, protein_file, temporal_file)
            filtered_file = method_output_file + '.filt'
            blast_filter_slow(temporal_file, filtered_file, id_perc, bitscore, evalue, aln_percent)
            Path(temporal_file).unlink()
            return (protein_file, filtered_file)









#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Parses compressed genbank files and extracts protein
information relevant for annotation purposes.
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from pathlib import Path
from typing import List
from Bio import SeqIO
from sys import argv

import multiprocessing
import argparse
import shutil
import gzip
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Function to parse Genbank files and create table
def table_creator(genbank_file: Path) -> Path:
    """
    Parses Genbank files and extract annotation information

    Args:
        genbank_file (Path): Path to compressed genbank file

    Returns:
        Path: Path to temporal table created
    """
    temporal_table = Path(genbank_file).with_suffix('.temp')
    with gzip.open(genbank_file, 'rt') as uncompressed_genbank, \
    open(temporal_table, 'w') as output_file:
        for record in SeqIO.parse(uncompressed_genbank, "genbank"):
            protein_id = record.id
            product = record.description.split("[")[0].strip()
            product = product.lower()
            taxonomy = ""
            species = ""
            ec_number = ""
            if 'organism' not in record.annotations:
                species = 'NA'
            else:
                species = record.annotations['organism']
                if 'taxonomy' not in record.annotations and species == "NA":
                    taxonomy = 'NA'
                elif 'taxonomy' not in record.annotations and species != "NA":
                    taxonomy = species
                else:
                    taxonomy = f"{', '.join(record.annotations['taxonomy'])}, {species}" 
                if product == "" or "unknown" in product:
                    for feature in record.features:
                        if feature.type == "Protein":
                            if 'product' not in feature.qualifiers:
                                if 'name' in feature.qualifiers:
                                    product = feature.qualifiers['name'][0]
                                else:
                                    product = 'NA'
                            else:
                                product = feature.qualifiers['product'][0]
            for feature in record.features:
                if feature.type == "Protein" and \
                    "EC_number" in feature.qualifiers:
                    ec_number = feature.qualifiers["EC_number"][0]
                else:
                    ec_number = "NA"
            output_file.write(
                f"{protein_id}\t{product}\t{taxonomy}\t{ec_number}\n")

    return temporal_table

# Function to parse Genbank files using multiprocessing
def table_generator_worker(
    genbank_list: List[Path], threads: int) -> List[Path]:
    """
    Worker calling table_creator function

    Args:
        genbank_list (List[Path]): List of compressed GenBank files
        threads (int): Threads to use

    Returns:
        List[Path]: List of temporal tables created
    """
    logger.info("Processing GenBank files")
    temp_table_list = []
    try:
        pool = multiprocessing.Pool(threads)
        temp_table_list = pool.map(table_creator, genbank_list)
    finally:
        pool.close()
        pool.join()
    logger.info("Finished")

    return temp_table_list

# Funtion to merge temporary files
def table_merger(temp_table_list: List[Path], final_table: Path) -> None:
    """
    Merges individual annotation tables into a single RefSeq annotation table

    Args:
        temp_table_list (List[Path]): List of paths to temporary tables created
        final_table (Path): Final table file
    """
    logger.info("Merging RefSeq annotation tables")
    with open(final_table, 'w') as output_file:
        for file in temp_table_list:
            with open(file) as temp:
                shutil.copyfileobj(temp, output_file)
    logger.info("Finished")
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script parses compressed RefSeq Genbank files,\n"
            f"extracts relevant information and stores in a single table.\n"
            f"Mandatory parameters: -i [input file(s)] -o [output table]\n"
            f"Optional parameters: See {argv[0]} -h\n"))
    # Setup mandatory arguments
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-i', '--input_files', dest='input_files', action='store',
        nargs='+', required=True,
        help=(
            f"Space-separated list of compressed genbank files,\n"
            f"e.g., '$(ls *.gz).'"))
    mandatory_arguments.add_argument(
        '-o', '--output_file', dest='output_file',
        action='store', required=True,
        help="File to store final table with record information.")
    # Setup optional arguments
    optional_arguments = parser.add_argument_group("Optional")
    optional_arguments.add_argument(
        '-t', '--threads', dest='threads', action='store',
        default=1, type=int, required=False,
        help='Number of threads to use, by default 1.')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    input_files = arguments.input_files
    output_file = arguments.output_file
    threads = arguments.threads
    # Run functions
    temp_tables = table_generator_worker(input_files, threads)
    table_merger(temp_tables, output_file)
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================



#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Parses compressed dat files and extracts protein
information relevant for annotation purposes.
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from pathlib import Path
from sys import argv

import argparse
import gzip
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Function to parse UniProt dat files
def parse_uniprot_dat(dat_file: Path, output_file_table: Path):
    output_folder = output_file_table.parent
    uniprot_to_refseq = Path(output_folder) / "uniprot_to_refseq.txt"
    logger.info(f"Parsing {dat_file} and storing in {output_file_table}")
    with gzip.open(dat_file, 'rt') as uniprot, \
        open(output_file_table, 'w') as output_file, \
        open(uniprot_to_refseq, 'a') as uni_to_ref:
        gene_id = ""
        accession = ""
        gene_name = ""
        ko_number = ""
        organism = ""
        taxonomy = ""
        function = ""
        compartment = ""
        process = ""
        interpro = ""
        pfam = ""
        ec_number = ""
        refseq_code = ""
        for line in uniprot:
            if line.startswith("ID", 0):
                gene_id = line.split()[1]
            elif "AC  " in line:
                accession = line.split()[1].replace(";","")
            elif line.startswith("DE") and "Full=" in line:
                gene_name = line.split("Full=")[1]
                gene_name = gene_name.split("{")[0].strip().replace(";","")
                gene_name = gene_name.lower()
            elif "OS  " in line:
                line = line.split("OS")[1].strip()
                organism = ' '.join([organism, line]).replace(".","")
            elif "OC  " in line:
                line = line.split("OC")[1].strip()
                taxonomy = ' '.join([taxonomy, line]).replace(".","")
            elif "DR   KO;" in line:
                ko_number = line.split()[2].replace(";", "")
            elif "DR   GO;" in line:
                if "; F:" in line:
                    code = line.strip().split(";")[1]
                    code = code.strip()
                    if function == "":
                        function = code
                    else:
                        function = ''.join([function, " ", code])
                elif "; C:" in line:
                    code = line.strip().split(";")[1]
                    code = code.strip()
                    if compartment == "":
                        compartment = code
                    else:
                        compartment = ''.join([compartment, " ", code])
                elif "; P:" in line:
                    code = line.strip().split(";")[1]
                    code = code.strip()
                    if process == "":
                        process = code
                    else:
                        process = ''.join([process, " ", code])
            elif "DR   InterPro" in line:
                code = line.strip().split()[2]
                code = code.replace(";", "")
                if interpro == "":
                    interpro = code
                else:
                    interpro = ' '.join([interpro, code])
            elif "DR   Pfam" in line:
                code = line.strip().split()[2]
                code = code.replace(";", "")
                if pfam == "":
                    pfam = code
                else:
                    pfam = ' '.join([pfam, code])
            elif line.startswith("DE") and "EC=" in line:
                ec_code = line.strip().split()[1]
                ec_code = ec_code.replace(";", "")
                ec_code = ec_code.replace("EC=", "")
                if ec_number == "":
                    ec_number = ec_code
                else:
                    ec_number = ' '.join([ec_number, ec_code])
            elif line.startswith("DR") and "RefSeq" in line:
                refseq_code = line.strip().split()[2]
                refseq_code = refseq_code.replace(";", "")
                uni_to_ref.write("{}\t{}\n".format(gene_id, refseq_code))
            elif "//\n" in line:
                if ko_number == "":
                    ko_number = "NA"
                if organism == "":
                    organism = "NA"
                if function == "":
                    function = "NA"
                if compartment == "":
                    compartment = "NA"
                if process == "":
                    process = "NA"
                if interpro == "":
                    interpro = "NA"
                if pfam == "":
                    pfam = "NA"
                if ec_number == "NA":
                    ec_number = "NA"
                output_file.write(
                    f"{gene_id}\t{accession}\t{gene_name}\t{ko_number}\t"
                    f"{organism}\t{taxonomy}\t{function}\t{compartment}\t"
                    f"{process}\t{interpro}\t{pfam}\t{ec_number}\n")
                gene_id = ""
                accession = ""
                gene_name = ""
                ko_number = ""
                organism = ""
                taxonomy = ""
                function = ""
                compartment = ""
                process = ""
                interpro = ""
                pfam =""
                ec_number = ""
                refseq_code = ""
    logger.info(f"Finished")

    return uniprot_to_refseq
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script parses a UniProt.dat file and returns a \n"
            f"tab-separated table with: \n"
            f"GeneID - Accession - Gene Name - KO Number - Organism - "
            f"Taxonomy - Function GO - Compartment GO - Process GO - "
            f"InterPro ID - Pfam ID - EC Number\n"
            f"Mandatory parameters: -i [input .dat file] -o [output table]\n"
            f"Optional parameters: See {argv[0]} -h\n"))
    # Setup mandatory arguments
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-i', '--input', dest='input_dat', action='store',
        required=True, help='Uniprot.dat file to parse')
    mandatory_arguments.add_argument(
        '-o', '--output_file', dest='output_file_table',action='store',
        required=True, help='output_file table')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    input_dat = arguments.input_dat
    output_file_table = arguments.output_file_table
    # Run functions
    parse_uniprot_dat(input_dat, output_file_table)
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================



#!/usr/bin/env python

"""
########################################################################
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Creates a table to KO to EC table based on KEGG information.
# This script requires you to download the 'htext' file from
# https://www.kegg.jp/kegg-bin/get_htext?ko00001
########################################################################
"""

################################################################################
"""---0.0 Import Modules---"""


################################################################################
"""---1.0 Define Functions---"""
def kegg_orthology_parser(kegg_file, outfile):
    ko_ec_dict= {}
    with open(kegg_file, 'r') as infile:
        for line in infile:
            if line.startswith('D') and "EC:" in line:
                ko_id = line.strip().split()[1]
                ec_numbers = line.strip().split("EC:")[1]
                ec_numbers = ec_numbers.replace("]", "")
                ec_numbers = ec_numbers.split()
                ko_ec_dict[ko_id] = ec_numbers
    with open(outfile, 'w') as output:
        for ko_identifier, ec_identifiers in ko_ec_dict.items():
            for element in ec_identifiers:
                output.write("{}\t{}\n".format(ko_identifier, element))

################################################################################
"""---2.0 Main Function---"""

def main():
    import argparse, sys
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
            description='''Creates a table to KO to EC table based on KEGG information.\n'''
            '''This script requires you to download the 'htext' file from\n'''
            '''https://www.kegg.jp/kegg-bin/get_htext?ko00001.\n'''
            '''Usage: ''' + sys.argv[0] + ''' -i [input_table] -o [output_table]\n'''
            '''Global mandatory parameters: -i [input_table] -o [output_table]\n'''
            '''Optional Database Parameters: See ''' + sys.argv[0] + ' -h')
    parser.add_argument('-i', '--input', dest='infile', action='store', required=False,
                        help='Input table downloaded from KEGG')
    parser.add_argument('-o', '--output', dest='output', action='store', required=False,
                        help='Output table with ko to ec ids.')
    args = parser.parse_args()

    infile = args.infile
    output = args.output

    # ----------------------------
    kegg_orthology_parser(infile, output)
    # ----------------------------

if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Date:         Nov 13, 2020

# Description: Download the latest refseq protein dataset from NCBI and
# the accompanying genbank files necessary to create the database.
# ==============================================================================
"""


# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from shutil import copyfileobj
from shutil import rmtree
from shutil import which
from pathlib import Path
from typing import List
from sys import argv

import multiprocessing
import urllib.request
import subprocess
import argparse
import gzip
import wget
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Function to download refseq data using aspera connect
def refseq_fasta_downloader(
    output_file_folder: Path, ascp_key: str = None) -> None:
    """
    Downloads all RefSeq proteins using Aspera connect
    and merges them in a single output

    Args:
        output_file_folder (Path): Folder to store downloaded protein files.
        ascp_key (str, optional): Aspera SSH key. Defaults to None.
    """
    logger.info("Downloading protein fasta files using Aspera Connect.")
    # Get aspera ssh key
    if ascp_key == None:
        ascp_key = get_aspera_key()
    # Create required folders
    temp_fasta_files = output_file_folder / "temp_refseq_proteins"
    merged_db_folder = output_file_folder / "protein_db"
    temp_fasta_files.mkdir(parents=True, exist_ok=True)
    merged_db_folder.mkdir(parents=True, exist_ok=True)
    # Get latest release for RefSeq and store it
    release_file = urllib.request.urlopen(
        "https://ftp.ncbi.nlm.nih.gov/refseq/release/RELEASE_NUMBER")
    release_number = release_file.readline().decode('utf-8').strip()
    with open(merged_db_folder / "refseq_release.txt", 'w') as relinfo:
        relinfo.write(f"Release number {release_number}")
    # Remove refseq fasta files if present (to avoid repeating files)
    if Path(merged_db_folder / "refseq_protein.fasta").is_file():
        Path(merged_db_folder / "refseq_protein.fasta").unlink()
    # Download compressed protein files
    subprocess.call(["ascp", '-QTr', '-d', '-l', '100M', '-E', '*wgs_mstr*',
                    '-E', '*rna*', '-E', '*genomic*', '-E', '*.gpff.gz',
                    '-i', ascp_key,
                    'anonftp@ftp.ncbi.nlm.nih.gov:/refseq/release/viral/',
                    temp_fasta_files])
    subprocess.call(["ascp", '-QTr', '-d', '-l', '100M', '-E', '*wgs_mstr*',
                    '-E', '*rna*', '-E', '*genomic*', '-E', '*.gpff.gz',
                    '-i', ascp_key,
                    'anonftp@ftp.ncbi.nlm.nih.gov:/refseq/release/bacteria/',
                    temp_fasta_files])
    subprocess.call(["ascp", '-QTr', '-d', '-l', '100M', '-E', '*wgs_mstr*',
                    '-E', '*rna*', '-E', '*genomic*', '-E', '*.gpff.gz',
                    '-i', ascp_key,
                    'anonftp@ftp.ncbi.nlm.nih.gov:/refseq/release/archaea/',
                    temp_fasta_files])
    # Get downloaded files
    viral_list = search_all_files(temp_fasta_files / "viral")
    bacteria_list = search_all_files(temp_fasta_files / "bacteria")
    archaea_list = search_all_files(temp_fasta_files / "archaea")
    final_list = viral_list + bacteria_list + archaea_list
    refseq_proteins = merged_db_folder / "refseq_protein.fasta"
    logger.info("Merging protein files")
    with open(refseq_proteins, 'w') as merged_db:
        for file in final_list:
            with gzip.open(file, 'rt') as temp_file:
                copyfileobj(temp_file, merged_db)
                Path.unlink(file)
    rmtree(temp_fasta_files)
    logger.info("RefSeq proteins successfully downloaded.")

    return refseq_proteins
    
def refseq_fasta_downloader_wget(
    output_file_folder: Path, threads: int) -> None:
    """
    Downloads all RefSeq proteins using wget
    and merges them in a single output

    Args:
        output_file_folder (Path): Folder to store downloaded protein files.
        threads (int): Number of threads to use.
    """
    logger.info("Downloading protein fasta files using wget.")
    # Create required folders
    temp_fasta_files = Path(output_file_folder) / "temp_refseq_proteins"
    merged_db_folder = Path(output_file_folder) / "protein_db"
    Path(temp_fasta_files).mkdir(parents=True, exist_ok=True)
    Path(merged_db_folder).mkdir(parents=True, exist_ok=True)
    # Get latest release for RefSeq and store it
    release_file = urllib.request.urlopen(
        "https://ftp.ncbi.nlm.nih.gov/refseq/release/RELEASE_NUMBER")
    release_number = release_file.readline().decode('utf-8').strip()
    with open(merged_db_folder / "refseq_release.txt", 'w') as relinfo:
        relinfo.write("Release number {}".format(release_number))
    # Remove refseq fasta files if present (to avoid repeating files)
    if Path(merged_db_folder / "refseq_protein.fasta").is_file():
        Path(merged_db_folder / "refseq_protein.fasta").unlink()
    # Download compressed protein files using wget
    file_list = []
    for domain in ["viral", "bacteria", "archaea"]:
        ncbi_url = "https://ftp.ncbi.nlm.nih.gov/refseq/release/" + domain
        domain_url_info = urllib.request.urlopen(ncbi_url)
        for line in domain_url_info:
            line = line.decode('utf-8').strip()
            if "protein.faa.gz" in line:
                filename = str(line.split('"')[1])
                file_url = (
                    f"ftp://ftp.ncbi.nlm.nih.gov/refseq/release/"
                    f"{domain}/{filename}"
                )
                output = f"{str(temp_fasta_files)}/{filename}"
                file_list.append((file_url, output))            
    try:
        pool = multiprocessing.Pool(threads)
        pool.map(refseq_multiprocess_downloader, file_list)
    finally:
        pool.close()
        pool.join()
    # Merge downloaded files
    logger.info("Merging protein files")
    final_list = search_all_files(temp_fasta_files)
    refseq_proteins = merged_db_folder / "refseq_protein.fasta"
    with open(refseq_proteins, 'w') as merged_db:
        for file in final_list:
            with gzip.open(file, 'rt') as temp_file:
                copyfileobj(temp_file,merged_db)
                Path.unlink(file)
    rmtree(temp_fasta_files)
    logger.info("RefSeq proteins successfully downloaded.")

    return refseq_proteins


def refseq_genbank_downloader(
    output_file_folder: Path,
    ascp_key: str = None) -> List[Path]:
    """
    Downloads RefSeq Genbank files with protein metadata using Aspera Connect.

    Args:
        output_file_folder (Path): Folder to store downloaded Genbank files.
        ascp_key (str, optional): Aspera SSH key. Defaults to None.

    Returns:
        List[Path]: List of Genbank files.
    """
    logger.info("Downloading protein genbank files using Aspera Connect.")
    # Get aspera ssh key
    if ascp_key == None:
        ascp_key = get_aspera_key()
    # Make output folder
    temp_gb_dir = output_file_folder / "temp_genbank"
    temp_gb_dir.mkdir(parents=True, exist_ok=True)
    # Download compressed protein files
    subprocess.call(["ascp", '-QTr', '-d', '-l', '100M', '-E', '*wgs_mstr*',
                    '-E', '*rna*', '-E', '*genomic*', '-E', '*.faa.gz',
                    '-i', ascp_key,
                    'anonftp@ftp.ncbi.nlm.nih.gov:/refseq/release/viral/',
                    temp_gb_dir])
    subprocess.call(["ascp", '-QTr', '-d', '-l', '100M', '-E', '*wgs_mstr*',
                    '-E', '*rna*', '-E', '*genomic*', '-E', '*.faa.gz',
                    '-i', ascp_key,
                    'anonftp@ftp.ncbi.nlm.nih.gov:/refseq/release/bacteria/',
                    temp_gb_dir])
    subprocess.call(["ascp", '-QTr', '-d', '-l', '100M', '-E', '*wgs_mstr*',
                    '-E', '*rna*', '-E', '*genomic*', '-E', '*.faa.gz',
                    '-i', ascp_key,
                    'anonftp@ftp.ncbi.nlm.nih.gov:/refseq/release/archaea/',
                    temp_gb_dir])
    # Get downloaded files
    viral_list = search_all_files(temp_gb_dir / "viral")
    bacteria_list = search_all_files(temp_gb_dir / "bacteria")
    archaea_list = search_all_files(temp_gb_dir / "archaea")
    final_list = viral_list + bacteria_list + archaea_list
    logger.info("RefSeq genbank successfully downloaded.")

    return temp_gb_dir, final_list

def refseq_genbank_downloader_wget(
    output_file_folder: Path, threads: int) -> List[Path]:
    """
    Downloads RefSeq Genbank files with protein metadata using wget.

    Args:
        output_file_folder (Path): [description]
        threads (int): [description]

    Returns:
        List[Path]: [description]
    """
    logger.info("Downloading protein genbank files using wget.")
    # Make output folder
    temp_gb_dir = output_file_folder / "temp_genbank"
    temp_gb_dir.mkdir(parents=True, exist_ok=True)
    # Download compressed genbank files using wget
    file_list = []
    for domain in ["viral", "bacteria", "archaea"]:
        ncbi_url = "https://ftp.ncbi.nlm.nih.gov/refseq/release/" + domain
        domain_url_info = urllib.request.urlopen(ncbi_url)
        for line in domain_url_info:
            line = line.decode('utf-8').strip()
            if "protein.gpff.gz" in line:
                filename = str(line.split('"')[1])
                file_url = (
                    f"ftp://ftp.ncbi.nlm.nih.gov/refseq/release/"
                    f"{domain}/{filename}"
                )
                output = f"{str(temp_gb_dir)}/{filename}"
                file_list.append((file_url, output))
    # Download using multiple cores
    try:
        pool = multiprocessing.Pool(threads)
        pool.map(refseq_multiprocess_downloader, file_list)
    finally:
        pool.close()
        pool.join()
    final_list = search_all_files(temp_gb_dir)
    logger.info("RefSeq genbank successfully downloaded.")

    return temp_gb_dir, final_list

def refseq_multiprocess_downloader(file_list: List[str]) -> None:
    """
    Downloads data from refseq using wget

    Args:
        file_list (List[Path]): List of files to download.
    """
    file_url = file_list[0]
    output = file_list[1]
    wget.download(file_url, out=output)

def get_aspera_key() -> str:
    """
    Retrieves Aspera Connect ssh key

    Returns:
        str: Path to Aspera Connect ssh key file
    """
    path_to_ascp = which("ascp")
    install_folder = Path(path_to_ascp).parents[1]
    key = f"{install_folder}/etc/asperaweb_id_dsa.openssh"
    
    return key

def search_all_files(directory: Path) -> List[Path]:
    """
    Search all compressed files in a directory

    Args:
        directory (Path): Directory to search.

    Returns:
        List[Path]: List of files ending in .gz in search directory.
    """
    dirpath = Path(directory)
    assert(dirpath.is_dir())
    file_list = []
    for file in dirpath.iterdir():
        if file.is_file() and file.suffix == ".gz":
            file_list.append(file)
        else:
            continue

    return file_list
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script downloads the Fasta and Genbank files that\n"
            f"are needed to build the RefSeq annotation database.\n"
            f"By default it downloads both but either can be specified.\n"
            f"Mandatory parameters: -o [output directory]\n"
            f"Optional parameters: See {argv[0]} -h\n"))
    # Setup mandatory arguments
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-o', '--outdir', dest='outdir', action='store', required=True,
        help='Output directory to store the Fasta and Genbank files.')
    # Setup optional arguments
    optional_arguments = parser.add_argument_group("Optional")
    optional_arguments.add_argument(
        '--proteins', dest='proteins', action='store_true', required=False,
        help='Only download the protein Fasta files.')
    optional_arguments.add_argument(
        '--genbank', dest='genbank', action='store_true', required=False,
        help='Only download the Genbank files.')
    optional_arguments.add_argument(
        '--ascp', dest='ascp_key', action='store', required=False, default=None,
        help='Path to Aspera Connect key.')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    outdir = arguments.outdir
    proteins = arguments.proteins
    genbank = arguments.genbank
    ascp_key = arguments.ascp_key

    # Run functions
    if proteins and genbank:
        logger.error(
            f"Please specify only either --proteins or --genbank.\n"
            f"If you want both do not pass any of these flags.")
    if proteins:
        refseq_fasta_downloader(outdir, ascp_key)
    elif genbank:
        refseq_genbank_downloader(outdir, ascp_key)
    else:
        refseq_fasta_downloader(outdir, ascp_key)
        refseq_genbank_downloader(outdir, ascp_key)
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================



#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Download the latest uniprot protein datasets from Uniprot and
# the accompanying .dat files necessary to create the database.
# ==============================================================================
"""


# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from shutil import copyfileobj
from pathlib import Path
from sys import argv

import urllib.request
import argparse
import wget
import gzip
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Function to download Swissprot protein sequences
def swissprot_fasta_downloader(database_directory: Path) -> Path:
    """
    Downloads protein sequences from UniProt SwissProt.

    Args:
        database_directory (Path): Database directory to store files. 

    Returns:
        Path: Fasta file with SwissProt protein sequences.
    """
    # Create protein folder
    merged_db_folder = database_directory / "protein_db"
    merged_db_folder.mkdir(parents=True, exist_ok=True)
    # Get release number
    rel_url = (
        f"ftp://ftp.uniprot.org/pub/databases/uniprot/"
        f"current_release/knowledgebase/complete/reldate.txt")
    release_file = urllib.request.urlopen(rel_url)
    release_number = release_file.readline().decode('utf-8').split()[3]
    # Write release information into file
    with open(merged_db_folder / "uniprot_release.txt", 'w') as relinfo:
        relinfo.write("Release number (date) {}".format(release_number))
    # Download fasta files
    logger.info("Downloading SwissProt Fasta files")
    # Overwrite existing temporal files
    output_fasta = merged_db_folder / "uniprot_sprot.fasta.gz"
    if output_fasta.is_file():
        output_fasta.unlink()
    # Download swissprot
    fasta_url = (
        f"ftp://ftp.uniprot.org/pub/databases/uniprot/"
        f"current_release/knowledgebase/complete/uniprot_sprot.fasta.gz")
    wget.download(fasta_url, out=str(output_fasta))
    decompressed_fasta = merged_db_folder / "uniprot_sprot.fasta"
    with gzip.open(output_fasta, 'rt') as compressed_fh, \
        open(decompressed_fasta, 'w') as decompressed_fh:
        copyfileobj(compressed_fh, decompressed_fh)
        output_fasta.unlink()
    logger.info("Finished")

    return output_fasta

# Function to download SwissProt annotations
def swissprot_dat_downloader(database_directory: Path) -> Path:
    """
    Downloads UniProt SwissProt protein annotations in the form of a .dat file.

    Args:
        database_directory (Path): Database directory to store files.

    Returns:
        Path: Path to downloaded .dat annotation file.
    """
    # Create dat files folder
    temp_dat_files = database_directory / "temp_swissprot_dat_files"
    temp_dat_files.mkdir(parents=True, exist_ok=True)
    # Overwrite existing temporal files
    output_dat_file = temp_dat_files / "uniprot_sprot.dat.gz"
    if output_dat_file.is_file():
        output_dat_file.unlink()
    # Download swissprot and trembl dat files
    logger.info("Downloading Swissprot .dat files")
    dat_url = (
        f"ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/"
        f"knowledgebase/complete/uniprot_sprot.dat.gz"
    )
    wget.download(dat_url, out=str(output_dat_file))
    logger.info("Finished")

    return output_dat_file

# Function to download TrEMBL protein sequences
def trembl_fasta_downloader(database_directory: Path) -> Path:
    """
    Downloads protein sequences from UniProt TrEMBL.

    Args:
        database_directory (Path): Database directory to store files. 

    Returns:
        Path: Fasta file with TrEMBL protein sequences.
    """
    # Create protein folder
    merged_db_folder = database_directory / "protein_db"
    merged_db_folder.mkdir(parents=True, exist_ok=True)
    # Get release number
    rel_url = (
        f"ftp://ftp.uniprot.org/pub/databases/uniprot/"
        f"current_release/knowledgebase/complete/reldate.txt")
    release_file = urllib.request.urlopen(rel_url)
    release_number = release_file.readline().decode('utf-8').split()[3]
    # Write release information into file
    with open(merged_db_folder / "uniprot_release.txt", 'w') as relinfo:
        relinfo.write("Release number (date) {}".format(release_number))
    # Download fasta files
    logger.info("Downloading TrEMBL Fasta files")
    # Overwrite existing temporal files
    output_fasta = merged_db_folder / "uniprot_trembl.fasta.gz"
    if output_fasta.is_file():
        output_fasta.unlink()
    # Download swissprot
    fasta_url = (
        f"ftp://ftp.uniprot.org/pub/databases/uniprot/"
        f"current_release/knowledgebase/complete/uniprot_trembl.fasta.gz")
    wget.download(fasta_url, out=str(output_fasta))
    decompressed_fasta = merged_db_folder / "uniprot_trembl.fasta"
    with gzip.open(output_fasta, 'rt') as compressed_fh, \
        open(decompressed_fasta, 'w') as decompressed_fh:
        copyfileobj(compressed_fh, decompressed_fh)
        output_fasta.unlink()
    logger.info("Finished")

# Function to download TrEMBL annotations
def trembl_dat_downloader(database_directory: Path) -> Path:
    """
    Downloads UniProt TrEMBL protein annotations in the form of a .dat file.

    Args:
        database_directory (Path): Database directory to store files.

    Returns:
        Path: Path to downloaded .dat annotation file.
    """
    # Create dat files folder
    temp_dat_files = database_directory / "temp_trembl_dat_files"
    temp_dat_files.mkdir(parents=True, exist_ok=True)
    # Overwrite existing temporal files
    output_dat_file = temp_dat_files / "uniprot_trembl.dat.gz"
    if output_dat_file.is_file():
        output_dat_file.unlink()
    # Download swissprot and trembl dat files
    logger.info("Downloading TrEMBL .dat files")
    dat_url = (
        f"ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/"
        f"knowledgebase/complete/uniprot_trembl.dat.gz"
    )
    wget.download(dat_url, out=str(output_dat_file))
    logger.info("Finished")

    return output_dat_file
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script downloads the Fasta and UniProt.dat files\n"
            f"that are needed to build the Uniprot annotation database.\n"
            f"By default it downloads both but either can be specified.\n"
            f"Mandatory parameters: -o [output directory]\n"
            f"Optional parameters: See {argv[0]} -h\n"))
    # Setup mandatory arguments
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-o', '--outdir', dest='outdir',
        action='store', required=True,
        help='Output directory to store the fasta and .dat files.')
    # Setup optional arguments
    optional_arguments = parser.add_argument_group("Optional")
    optional_arguments.add_argument(
        '--light', dest='light', action='store_true', required=False,
        help='Light mode, only downloads the swissprot-related data.')
    optional_arguments.add_argument(
        '--proteins', dest='proteins', action='store_true', required=False,
        help='Only download the protein Fasta files.')
    optional_arguments.add_argument(
        '--datfile', dest='datfile', action='store_true', required=False,
        help='Only download the .dat files.')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    outdir = arguments.outdir
    proteins = arguments.proteins
    datfile = arguments.datfile
    light = arguments.light

    # Run functions
    if light:
        if proteins:
            swissprot_fasta_downloader(outdir)
        elif datfile:
            swissprot_dat_downloader(outdir)
        else:
            swissprot_fasta_downloader(outdir)
            swissprot_dat_downloader(outdir)
    else:
        if proteins:
            swissprot_fasta_downloader(outdir)
            trembl_fasta_downloader(outdir)
        elif datfile:
            swissprot_dat_downloader(outdir)
            trembl_dat_downloader(outdir)
        else:
            swissprot_fasta_downloader(outdir)
            trembl_fasta_downloader(outdir)
            swissprot_dat_downloader(outdir)
            trembl_dat_downloader(outdir)
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================



#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz-Perez
# Email:        cruizperez3@gatech.edu
# Institution:  Georgia Institute of Technology
# Date:         11 April 2021

# Description: This script contains utilities used by scripts in several
# locations.
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from random import choices
from shutil import which
from pathlib import Path
from typing import List
from sys import exit

import psutil
import string
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions 
# ==============================================================================
# Generate random strings
def random_string(length: int) -> str:
    """
    Generate random string

    Args:
        length (int): Lenght of the string.

    Returns:
        str: Random string of length [lenght]
    """
    return ''.join(choices(string.ascii_uppercase + string.digits, k = length))

# Validate input parameters
def input_validator(
    method: str, method_bin: Path, input_list: List[str], file_list: Path,
    cluster: str, processes: int, threads: int):
    # Validate method
    method_options = ['blast', 'sword', 'diamond']
    if method not in method_options:
        logger.error(
            f"Invalid search method: {method}.\n"
            f"Please select one of 'diamond' 'blast', or 'sword'")
        exit(1)
    # Validate method call
    call_name = None
    if method == 'blast':
        call_name = 'blastp'
    elif method == 'diamond':
        call_name = 'diamond'
    elif method == 'sword':
        call_name = 'sword'
    if method_bin == None:
        method_call = call_name
    else:
        method_call = str(method_bin / call_name)
    if which(method_call) == None:
            if method_bin == None:
                logger.error(
                    f"{method_call} not found in PATH.\n"
                    f"Please provide the correct path to the "
                    f"directory with {method_call} binaries.")
                exit(1)
            else:
                logger.error(
                    f"{method_call} not found in {method_bin}.\n"
                    f"Please provide the correct path to the "
                    f"directory with {method_call} binaries.")
                exit(1)
    # Validate input files
    if input_list != None and file_list != None:
        logger.error(
            f"Please provide only a list of files (--input) or an\n"
            f"input list (--list), not both.\n")
        exit(1)
    elif input_list != None:
        for file in input_list:
            if not Path(file).is_file():
                logger.error(
                    f"File {file} does not exist. Please check"
                    f"the file exists and can be accessed."
                )
                exit(1)
    elif file_list != None:
        with open(file_list, 'r') as infile:
            for line in infile:
                if not Path(line.strip()).is_file():
                    logger.error(
                        f"File {line.strip()} does not exist. Please check"
                        f"the file exists and can be accessed."
                    )
                    exit(1)
    # Validate number of processes and threads
    max_cores = psutil.cpu_count()
    if (processes * threads) > max_cores:
        logger.warning(
            f"The number of processes {processes} and threads\n"
            f"{threads} exceeds the maximum number of cores\n"
            f"available in the system {max_cores}. This can result\n"
            f"in a decrease in performance.")
    # Validate cluster request
    cluster_options = ['cols', 'rows', 'both']
    if cluster != None and cluster not in cluster_options:
        logger.error(
            f"Invalid cluster option: {cluster}.\n"
            f"Please select one of 'cols' 'rows', or 'both'\n"
            f"or don't pass this flag for no clustering.")
        exit(1)
    
# ==============================================================================


#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Create SQLite database from a tab-delimited tables.
# Used to create annotation parsing databases.
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from pathlib import Path
from sys import exit
from sys import argv

import argparse
import sqlite3
import wget
import gzip
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Function to create SwissProt table
def create_swissprot_table(database: str, final_table: Path) -> None:
    """
    Creates and adds SwissProt table to SQLite database.

    Args:
        database (str): Path to SQLite database.
        final_table (Path): Table to read and add to database.
    """
    logger.info("Creating RefSeq table")
    conn = sqlite3.connect(database)
    cursor = conn.cursor()
    cursor.execute("DROP TABLE IF EXISTS swissprot")
    cursor.execute('CREATE TABLE swissprot \
        (gene_id TEXT, accession TEXT, gene_name TEXT, ko_number TEXT, \
        organism TEXT, taxonomy TEXT, function TEXT, compartment TEXT, \
        process TEXT, interpro TEXT, pfam TEXT, ec_number TEXT)')
    with open(final_table, 'r') as swissprot:
        record_counter = 0
        records = []
        for line in swissprot:
            # Commit changes after 500000 records
            if record_counter == 500000:
                cursor.execute("begin")
                cursor.executemany('INSERT INTO swissprot VALUES(?, ?, ?, ?, \
                    ?, ?, ?, ?, ?, ?, ?, ?)', records)
                cursor.execute("commit")
                record_counter = 0
                records = []
            else:
                records.append(tuple(line.rstrip('\n').split("\t")))
                record_counter += 1
        # Commit remaining records
        if record_counter > 0:
            cursor.execute("begin")
            cursor.executemany('INSERT INTO swissprot VALUES(?, ?, ?, ?, ?, \
                ?, ?, ?, ?, ?, ?, ?)', records)
            cursor.execute("commit")
        # Create index for faster access
        cursor.execute('CREATE INDEX swissprot_id ON swissprot (gene_id)')
    logger.info("Finished")

def create_trembl_table(database: str, final_table: Path) -> None:
    """
    Creates and adds TrEMBL table to SQLite database.

    Args:
        database (str): Path to SQLite database.
        final_table (Path): Table to read and add to database.
    """
    logger.info("Creating TrEMBL table")
    conn = sqlite3.connect(database)
    cursor = conn.cursor()
    cursor.execute("DROP TABLE IF EXISTS trembl")
    cursor.execute('CREATE TABLE trembl \
        (gene_id TEXT, accession TEXT, gene_name TEXT, ko_number TEXT, \
        organism TEXT, taxonomy TEXT, function TEXT, compartment TEXT, \
        process TEXT, interpro TEXT, pfam TEXT, ec_number)')
    with open(final_table, 'r') as trembl:
        record_counter = 0
        records = []
        for line in trembl:
            # Commit changes after 500000 records
            if record_counter == 500000:
                cursor.execute("begin")
                cursor.executemany('INSERT INTO trembl VALUES(?, ?, ?, ?, ?, \
                    ?, ?, ?, ?, ?, ?, ?)', records)
                cursor.execute("commit")
                record_counter = 0
                records = []
            else:
                records.append(tuple(line.rstrip('\n').split("\t")))
                record_counter += 1
        # Commit remaining records
        if record_counter > 0:
            cursor.execute("begin")
            cursor.executemany('INSERT INTO trembl VALUES(?, ?, ?, ?, ?, \
                ?, ?, ?, ?, ?, ?, ?)', records)
            cursor.execute("commit")
        # Create index for faster access
        cursor.execute('CREATE INDEX trembl_id ON trembl (gene_id)')
    logger.info("Finished")

def create_refseq_table(database: str, final_table: Path) -> None:
    """
    Creates and adds RefSeq table to SQLite database.

    Args:
        database (str): Path to SQLite database.
        final_table (Path): Table to read and add to database.
    """
    # Store protein ko annotations
    # Download refseq KO annotations
    logger.info("Parsing RefSeq KO numbers")
    annot_url = (
        f"http://enve-omics.ce.gatech.edu/data/public_microbeannotator/"
        f"01.RefSeq_KEGG_Annotations.txt.gz")
    annotation_file = wget.download(
        annot_url, out="refseq_ko_annotations.tar.gz")
    protein_annotations = {}
    with gzip.open(annotation_file, 'rt') as annotations:
        for line in annotations:
            line = line.rstrip('\n').split("\t")
            protein_annotations[line[0]] = [line[1], line[2]]
    Path(annotation_file).unlink()
    logger.info("Finished")
    # Connect to db and create table
    logger.info("Creating RefSeq table")
    conn = sqlite3.connect(database)
    cursor = conn.cursor()
    cursor.execute("DROP TABLE IF EXISTS refseq")
    cursor.execute('CREATE TABLE refseq \
        (gene_id TEXT, product TEXT, taxonomy TEXT, \
        ec_number TEXT, ko_number TEXT, ko_product TEXT)')
    with open(final_table, 'r') as refseq:
        record_counter = 0
        records = []
        for line in refseq:
            # Commit changes after 500000 records
            if record_counter == 500000:
                cursor.execute("begin")
                cursor.executemany(
                    'INSERT INTO refseq VALUES(?, ?, ?, ?, ?, ?)', records)
                cursor.execute("commit")
                record_counter = 0
                records = []
            else:
                line = line.rstrip('\n').split("\t")
                if line[0] in protein_annotations:
                    ko_number = protein_annotations[line[0]][0]
                    ko_product = protein_annotations[line[0]][1]
                    records.append((line[0], line[1], line[2], 
                        line[3], ko_number, ko_product))
                    record_counter += 1
                else:
                    records.append((line[0], line[1], line[2],
                        line[3], "NA", "NA"))
                    record_counter += 1
        # Commit remaining records
        if record_counter > 0:
            cursor.execute("begin")
            cursor.executemany(
                'INSERT INTO refseq VALUES(?, ?, ?, ?, ?, ?)', records)
            cursor.execute("commit")
        # Create index for faster access
        cursor.execute('CREATE INDEX refseq_id ON refseq (gene_id)')
    logger.info("Finished")
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script adds tab-separated tables to a SQLite database.\n"
            f"By default it assumes the first line of the input table has\n"
            f"headers. If not present, they must be provided using the\n"
            f"-- header flag as --header header1,header2,header3, etc."
            f"Mandatory parameters: -i [input table] -d [sqlite database]\n"
            f"-t [table type]\n"))
    # Setup mandatory arguments
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-i', '--input', dest='input_table', action='store', required=True,
        help='Input tab-delimited table to parse, by default assumes headers are present')
    mandatory_arguments.add_argument(
        '-d', '--database', dest='database', action='store', required=True,
        help='Database name, can be exisiting or new.')
    mandatory_arguments.add_argument(
        '-t', '--type', dest='db_type', action='store', required=True,
        help='Table type, can be one of swissprot, trembl, or refseq.')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    input_table = arguments.input_table
    database = arguments.database
    db_type = arguments.db_type
    db_type = db_type.lower()
    # Validate requested type
    type_choice = ['swissprot', 'trembl', 'refseq']
    if db_type not in type_choice:
        logger.error(f"Unvalid database type requested, {db_type}.")
        exit(1)
    # Run functions
    logger.info("Building database table")
    if db_type == "swissprot":
        create_swissprot_table(database, input_table)
    elif db_type == "trembl":
        create_trembl_table(database, input_table)
    elif db_type == "refseq":
        create_refseq_table(database, input_table)
    else:
        exit("Please specify a table type, swissprot, trembl or refseq")
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================



from setuptools import setup, find_packages
from microbeannotator import version
import os

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

setup(
    name='microbeannotator',
    version=version,

    scripts=['bin/microbeannotator', 'bin/microbeannotator_db_builder'],
    include_package_data=True,

    packages=find_packages(),

    python_requires='>=3.6, <=3.8',
    author='Carlos A. Ruiz Perez',
    author_email='cruizperez3@gatech.edu',
    description='A user friendly microbe genome annotation tool',
    license='Artistic License 2.0',
    keywords = ['genome annotation', 'protein', 'comparative genomics'],
    url='https://github.com/cruizperez/MicrobeAnnotator'
)



#!/usr/bin/env python

"""
# ==============================================================================
# Author:	   Carlos A. Ruiz-Perez
# Email:       cruizperez3@gatech.edu
# Institution: Georgia Institute of Technology
# Version:	   0.9
# Date:		   14 February 2020

# Description: This script runs hmmsearch and filters the results based on the
# thresholds described by kofamscan.
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.core_utils import random_string
from microbeannotator.utilities.logging import setup_logger
from microbeannotator.database import kofam_profile_parser as kofam

from tempfile import NamedTemporaryFile
from tempfile import TemporaryDirectory
from pathlib import Path
from typing import List
from typing import Tuple

import multiprocessing
import itertools
import hmmer
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions 
# ==============================================================================
def hmmer_search(protein_model: Tuple[Path, Path]) -> List[hmmer.tbl.TBLRow]:
    # Get protein and model paths
    protein_path = protein_model[0]
    model_path = protein_model[1]
    # Create model database
    model_db = hmmer.HMMER(model_path)
    # Create temporary directory
    temp_dir = TemporaryDirectory(suffix='_microbeannotator')
    temp_file = NamedTemporaryFile(suffix=random_string(20))
    temporal_result = Path(temp_dir.name) / Path(temp_file.name)
    search_result = model_db.search(
        protein_path,
        output=temporal_result)
    if len(search_result.tbl) > 0:
        search_result_filt = hmmer_filter(search_result.tbl)
    
        return search_result_filt


def hmmer_filter(
    hmmsearch_result: List[hmmer.tbl.TBLRow]) -> List[hmmer.tbl.TBLRow]:
    # Initialize list with filtered results
    hmmsearch_result_filt = []
    # Get model name, threshold and score_type
    model_name = hmmsearch_result[0].query.name
    threshold = hmm_model_info[model_name].threshold
    score_type = hmm_model_info[model_name].score_type
    if score_type == 'full':
        for result in hmmsearch_result:
            if float(result.full_sequence.score) >= threshold:
                hmmsearch_result_filt.append(result)
    elif score_type == 'domain':
        for result in hmmsearch_result:
            if float(result.best_1_domain.score) >= threshold:
                hmmsearch_result_filt.append(result)
    else:
        for result in hmmsearch_result:
            hmmsearch_result_filt.append(result)

    return hmmsearch_result_filt


def write_hmmsearch_results(
    combined_hmmsearch_results: List[hmmer.tbl.TBLRow],
    out_path: Path) -> None:
    with open(out_path, 'w') as output:
        output.write((
            "#\t\t\t\t\t----\tfull_sequence\t----\t----\tbest_1_domain\t----\t"
            "---\t---\t---\tdomain\tnumber\t"
            "estimation\t---\t---\tdescription_of_target\n"
            "# query_name\taccession\ttarget_name\taccession\tscore_type\t"
            "E-value\tscore\tbias\tE-value\tscore\tbias\texp\treg\tclu\tov\t"
            "env\tdom\trep\tinc\tdescription_of_target\n"
            "# -----------\t---------\t----------\t---------\t----------\t"
            "-------\t-----\t----\t-------\t-----\t----\t---\t---\t---\t--"
            "\t---\t---\t---\t---\t---------------------\n"))
        for result in combined_hmmsearch_results:
            if result == None:
                continue
            elif len(result) == 0:
                continue
            else:
                for record in result:
                    output.write((
                        f"{record.target.name}\t{record.target.accession}\t"
                        f"{record.query.name}\t{record.query.accession}\t"
                        f"{hmm_model_info[record.query.name].score_type}\t"
                        f"{record.full_sequence.e_value}\t"
                        f"{record.full_sequence.score}\t"
                        f"{record.full_sequence.bias}\t"
                        f"{record.best_1_domain.e_value}\t"
                        f"{record.best_1_domain.score}\t"
                        f"{record.best_1_domain.bias}\t"
                        f"{record.domain_numbers.exp}\t"
                        f"{record.domain_numbers.reg}\t"
                        f"{record.domain_numbers.clu}\t"
                        f"{record.domain_numbers.ov}\t"
                        f"{record.domain_numbers.env}\t"
                        f"{record.domain_numbers.dom}\t"
                        f"{record.domain_numbers.rep}\t"
                        f"{record.domain_numbers.inc}\t"
                        f"{hmm_model_info[record.query.name].definition}\n"))


def best_match_selector(raw_results: Path) -> Path:
    outfile = raw_results.parent / (raw_results.name + '.filt')
    with open(raw_results, 'r') as infile, \
        open(outfile, 'w') as output:
        best_matches = {}
        for line in infile:
            if line.startswith('#'):
                output.write(line)
            else:
                record = line.strip().split('\t')
                if record[0] not in best_matches:
                    best_matches[record[0]] = [record[6], line]
                else:
                    if record[6] > best_matches[record[0]][0]:
                        best_matches[record[0]] = [record[6], line]
        for record in best_matches.values():
            output.write(record[1])
    return outfile


def protein_to_model_mapper(
    protein_list: List[Path],
    model_list: List[Path]) -> List[Tuple[Path, Path]]:
    combined_list = []
    for path_product in itertools.product(*[protein_list, model_list]):
        combined_list.append(path_product)
    
    return combined_list


def write_hmmsearch_annotation(
    hmmsearch_results: Path, annotation_file: Path):
    with open(hmmsearch_results) as infile, \
        open(annotation_file, 'a') as output:
        for line in infile:
            if line.startswith("#"):
                continue
            else:
                line = line.strip().split('\t')
                query_id = line[0]
                ko_id = line[2]
                ko_product = line[19]
                output.write(
                    f"{query_id}\tNA\tNA\t{ko_id}\t{ko_product}\tNA\t"
                    f"NA\tNA\tNA\tNA\tNA\tNA\tkofam\n"
                )


def hmmsearch_launcher(
    protein_model_list: List[Tuple[Path, Path]],
    ko_list_db: Path,
    output_path: Path,
    processes: int,
    threads: int):
    # Launch hmmsearch in one process per protein file vs model
    global hmm_model_info
    try:
        hmm_model_info
    except NameError:
        hmm_model_info = kofam.parse_profiles(ko_list_db)
    try:
        pool_search = multiprocessing.Pool(processes * threads)
        combined_hmmsearch_results = pool_search.map(
            hmmer_search, protein_model_list)
    finally:
        pool_search.close()
        pool_search.join()
    
    logger.info("Filtering and saving results...")
    # Get results and write them to output file
    write_hmmsearch_results(combined_hmmsearch_results, output_path)
    # Read intermediate results and select best match per protein
    best_match_file = best_match_selector(output_path)
    # Get annotated and hypothetical proteins
    annotated_proteins = []
    hypothetical_proteins = []
    with open(best_match_file) as infile:
        for line in infile:
            if line.startswith('#'):
                continue
            else:
                line = line.strip().split('\t')
                if "hypothetical" not in line[19].lower() and \
                    "uncharacterized" not in line[19].lower() and \
                    "unknown function" not in line[10].lower():
                    annotated_proteins.append(line[0])
                else:
                    hypothetical_proteins.append(line[0])
    logger.info("Search against KOfam done!")
    return best_match_file, annotated_proteins, hypothetical_proteins



#!/usr/bin/env python

"""
# ==============================================================================
# Author:       Carlos A. Ruiz Perez
# Email:        cruizperez3@gatech.edu
# Intitution:   Georgia Institute of Technology
# Version:      1.0.0
# Date:         Nov 13, 2020

# Description: Builds the search databases depending on the search method
# selected. Can be one of blast, sword or diamond. These must be in the PATH
# or the path must be provided.
# ==============================================================================
"""

# ==============================================================================
# Import modules
# ==============================================================================
from microbeannotator.utilities.logging import setup_logger

from pathlib import Path
from shutil import which
from sys import exit
from sys import argv

import subprocess
import argparse
# ==============================================================================


# ==============================================================================
# Initalize logger
# ==============================================================================
logger = setup_logger(__name__)
# ==============================================================================


# ==============================================================================
# Define functions
# ==============================================================================
# Function to create blast databases
def blastp_db_creator(input_directory: Path, bin_path: Path = None) -> None:
    logger.info('Building blast databases')
    database_files = []
    if bin_path != None:
        makeblasdb_call = Path(bin_path) / 'makeblastdb'
    else:
        makeblasdb_call = 'makeblastdb'
    if which(makeblasdb_call) == None:
        logger.error(
            f"BLAST's binary {makeblasdb_call} not found.\n"
            f"Plase make sure BLAST is installed and if possible in PATH.\n"
        )
        exit(1)
    else:
        for protein_file in Path(input_directory).iterdir():
            if protein_file.suffix == '.fasta':
                output_file_name = protein_file.with_suffix("")
                database_files.append(output_file_name)
                subprocess.call(
                    [makeblasdb_call, '-in', protein_file, '-dbtype', 'prot',
                    '-out', output_file_name])
    with open(Path(input_directory)/'blast_db.list', 'w') as db_list:
        for file in database_files:
            db_list.write(f"{file.name}\n")
    logger.info('Finished')

# Function to create diamond databases
def diamond_db_creator(input_directory, threads, bin_path=None):
    logger.info('Building diamond databases')
    database_files = []
    if bin_path != None:
        diamond_call = Path(bin_path) / 'diamond'
    else:
        diamond_call = 'diamond'
    if which(diamond_call) == None:
        logger.error(
            f"Diamond's binary {diamond_call} not found.\n"
            f"Plase make sure Diamond is installed and if possible in PATH.\n"
        )
        exit(1)
    else:
        for protein_file in Path(input_directory).iterdir():
            if protein_file.suffix == '.fasta':
                output_file_name = protein_file.with_suffix("")
                final_db_name = Path(output_file_name).with_suffix('.dmnd')
                database_files.append(final_db_name)
                subprocess.call(
                    [diamond_call, 'makedb', '--in', protein_file, '-d',
                    output_file_name, '--threads', str(threads)])
    with open(Path(input_directory)/'diamond_db.list', 'w') as db_list:
        for file in database_files:
            db_list.write(f"{file.name}\n")
    logger.info('Finished')

# Function to create sword database
def sword_db_creator(input_directory, bin_path=None):
    logger.info('Building sword databases')
    logger.info('No DB needed for sword, just checking if sword is in PATH')
    if bin_path != None:
        sword_call = Path(bin_path) / 'sword'
    else:
        sword_call = 'sword'
    if which(sword_call) == None:
        logger.warning(
            f"Sword's binary {sword_call} not found.\n"
            f"Plase make sure Sword is installed and if possible in PATH.\n"
        )
    with open(Path(input_directory)/'sword_db.list', 'w') as db_list:
        for file in Path(input_directory).iterdir():
            if file.suffix == '.fasta':
                db_list.write(f"{file.name}\n")
    logger.info('Finished')
# ==============================================================================


# ==============================================================================
# Define main function
# ==============================================================================
def main():
    # Setup parser for arguments.
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"This script creates the search databases used by tools\n"
            f"in MicrobeAnnotator.\n"
            f"Mandatory parameters: -d [directory proteins] -m [method/tool]\n"
            f"-d [database directory] -s [sqlite database]\n"
            f"Optional parameters: See {argv[0]} -h"))
    # Setup mandatory arguments
    mandatory_arguments = parser.add_argument_group("Mandatory")
    mandatory_arguments.add_argument(
        '-d', '--dirprot', dest='dirprot', action='store', required=True,
        help='Directory where all raw fasta files are located.')
    mandatory_arguments.add_argument(
        '-m', '--method', dest='method', action='store', required=True,
        help='Search (and DB creation) method. One of blast, diamond or sword')
    # Setup optional arguments
    optional_arguments = parser.add_argument_group("Optional")
    optional_arguments.add_argument(
        '--bin_path', dest='bin_path', action='store', required=False,
        help='Path to binary folder for selected method/tool.')
    # If no arguments are provided
    if len(argv) == 1:
        parser.print_help()
        exit(0)
    arguments = parser.parse_args()

    # Parse arguments
    dirprot = arguments.dirprot
    method = arguments.method
    bin_path = arguments.bin_path
    method = method.lower()
    # Run functions
    if method == 'blast':
        blastp_db_creator(dirprot, bin_path)
    elif method == 'diamond':
        diamond_db_creator(dirprot, bin_path)
    elif method == 'sword':
        sword_db_creator(dirprot, bin_path)
    else:
        logger.error(
            f"Search method not recognized. Must be blast, diamond, or sword.")
        exit(1)
# ==============================================================================


# ==============================================================================
# Run main function
# ==============================================================================
if __name__ == "__main__":
    main()
# ==============================================================================
